# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys
import re
import typing

from cogment.errors import CogmentGenerateError

try:
    import click
    import yaml
except ModuleNotFoundError:
    raise CogmentGenerateError("The `cogment.generate` module requires extra dependencies, "
                       "please install by running `pip install cogment[generate]`")

ACTOR_CLASS_FORMAT = """
_{0}_class = _cog.actor.ActorClass(
            name="{0}",
            config_type={1},
            action_space={2},
            observation_space={3},
            )
"""

COG_SETTINGS_FORMAT = """
import cogment as _cog
from types import SimpleNamespace

{0}
{1}

actor_classes = _cog.actor.ActorClassList({2})

trial = SimpleNamespace(config_type={3})

environment = SimpleNamespace(config_type={4})
"""


def py_path_from_proto_path(proto_file_name: str) -> str:
    return proto_file_name.split(".")[0] + "_pb2"


def py_alias_from_proto_path(proto_file_name: str) -> str:
    return proto_file_name.split(".")[0] + "_pb"


def token_as_proto_token(yaml_token: str, proto_files_content) -> str:
    token_parts = yaml_token.split(".")
    if len(token_parts) == 1:
        token_package = ""
        token_name = token_parts[0]
    elif len(token_parts) == 2:
        token_package, token_name = token_parts
    else:
        raise CogmentGenerateError(f"Invalid token format [{yaml_token}]")

    if token_package not in proto_files_content:
        raise CogmentGenerateError(f"Unknown token package [{yaml_token}]")

    for proto_file_name, proto_file_data in proto_files_content[token_package]:
        # TODO: Improve robustness of this (e.g. if name is found even in a comment, it will "work")!
        if token_name in proto_file_data:
            return f"{py_alias_from_proto_path(proto_file_name)}.{token_name}"

    raise CogmentGenerateError(f"Token [{yaml_token}] not found in any file")


def proto_import_line(proto: str) -> str:
    return (f"import {py_path_from_proto_path(proto)} as {py_alias_from_proto_path(proto)}")


def proto_lib_line(proto: str) -> str:
    return f'protolib = "{py_path_from_proto_path(proto)}"'


def actor_class_line(actor_class) -> str:
    actor_className = actor_class["name"]
    return f"_{actor_className}_class"


def generate(spec: str, output: str):
    output_directory = os.path.dirname(os.path.abspath(output))
    spec_directory = os.path.dirname(os.path.abspath(spec))

    proto_files_content: typing.Dict[str, typing.List[typing.Tuple[str, str]]] = {}

    def actor_classes_block(actor_class) -> str:
        if actor_class["observation"].get("delta") is not None:
            raise CogmentGenerateError(f"[{actor_class['name']}] defines 'observation.delta' "
                                        f"which is no longer supported.")
        if actor_class["observation"].get("delta_apply_fn") is not None:
            raise CogmentGenerateError(f"[{actor_class['name']}] defines 'observation.delta_apply_fn' "
                                        f"which is no longer supported.")

        if "config_type" in actor_class:
            config_type = token_as_proto_token(actor_class["config_type"], proto_files_content)
        else:
            config_type = "None"

        return ACTOR_CLASS_FORMAT.format(
            actor_class["name"],
            config_type,
            token_as_proto_token(actor_class["action"]["space"], proto_files_content),
            token_as_proto_token(actor_class["observation"]["space"], proto_files_content)
        )

    with open(spec, "r") as stream:
        cog_settings = yaml.safe_load(stream)

        proto_files = cog_settings["import"]["proto"]
        package_pattern = re.compile(r"^\s*package\s+(\w+)\s*;", re.MULTILINE)

        for file_name in proto_files:
            with open(os.path.join(spec_directory, file_name), "r") as file:
                file_data = file.read()
                package = re.findall(package_pattern, file_data)
                if len(package) == 1:
                    package_name = package[0]
                elif len(package) == 0:
                    package_name = ""
                else:
                    raise CogmentGenerateError(f"Only one package can be defined in proto file [{file_name}]")

                if package_name not in proto_files_content:
                    proto_files_content[package_name] = [(file_name, file_data)]
                else:
                    proto_files_content[package_name].append((file_name, file_data))

        args = [
            sys.executable,
            "-m", "grpc_tools.protoc",
            "-I", ".",
            f"--python_out={output_directory}"
        ]
        args.extend(proto_files)

        subprocess.check_call(args, cwd=spec_directory)

        if len(cog_settings.get("import", {}).get("python", [])) > 0:
            raise CogmentGenerateError("Python imports in configuration file are not supported in Cogment 2.0 "
                                        "spec file")

        if "trial" in cog_settings and "config_type" in cog_settings["trial"]:
            trial_config = token_as_proto_token(cog_settings["trial"]["config_type"], proto_files_content)
        else:
            trial_config = "None"

        if "environment" in cog_settings and "config_type" in cog_settings["environment"]:
            env_config = token_as_proto_token(cog_settings["environment"]["config_type"], proto_files_content)
        else:
            env_config = "None"

        line_break = "\n"  # f strings can't have backslashes in them
        comma_line_break_tab = ",\n\t"  # f strings can't have backslashes in them

        proto_import_lines = []
        for proto_import_file in proto_files:
            proto_import_lines.append(proto_import_line(proto_import_file))

        actor_class_lines = []
        actor_class_blocks = []
        actor_classes = cog_settings.get("actor_classes", [])
        for actor_class in actor_classes:
            actor_class_lines.append(actor_class_line(actor_class))
            actor_class_blocks.append(actor_classes_block(actor_class))

        cog_settings_string = COG_SETTINGS_FORMAT.format(
            line_break.join(proto_import_lines),
            line_break.join(actor_class_blocks),
            comma_line_break_tab.join(actor_class_lines),
            trial_config,
            env_config)

        with open(output, "w") as text_file:
            text_file.write(cog_settings_string)


@click.command()
@click.option("--spec", default="cogment.yaml", help="Cogment spec file")
@click.option("--output", default="cog_settings.py", help="Output python file")
def main(spec: str, output: str):
    generate(spec, output)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x63ogment/api/common.proto\x12\ncogmentAPI\x1a\x19google/protobuf/any.proto\"\x10\n\x0eVersionRequest\"j\n\x0bVersionInfo\x12\x31\n\x08versions\x18\x01 \x03(\x0b\x32\x1f.cogmentAPI.VersionInfo.Version\x1a(\n\x07Version\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"$\n\x11SerializedMessage\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"9\n\rDatalogParams\x12\x10\n\x08\x65ndpoint\x18\x01 \x01(\t\x12\x16\n\x0e\x65xclude_fields\x18\x02 \x03(\t\"z\n\x11\x45nvironmentParams\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\x12\x16\n\x0eimplementation\x18\x03 \x01(\t\x12-\n\x06\x63onfig\x18\x04 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\"\x90\x02\n\x0b\x41\x63torParams\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x61\x63tor_class\x18\x02 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x03 \x01(\t\x12\x16\n\x0eimplementation\x18\x04 \x01(\t\x12-\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\x12\"\n\x1ainitial_connection_timeout\x18\x06 \x01(\x02\x12\x18\n\x10response_timeout\x18\x07 \x01(\x02\x12\x10\n\x08optional\x18\x08 \x01(\x08\x12\x35\n\x0e\x64\x65\x66\x61ult_action\x18\t \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\"\x91\x02\n\x0bTrialParams\x12\x33\n\x0ctrial_config\x18\x01 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\x12*\n\x07\x64\x61talog\x18\x02 \x01(\x0b\x32\x19.cogmentAPI.DatalogParams\x12\x32\n\x0b\x65nvironment\x18\x03 \x01(\x0b\x32\x1d.cogmentAPI.EnvironmentParams\x12\'\n\x06\x61\x63tors\x18\x04 \x03(\x0b\x32\x17.cogmentAPI.ActorParams\x12\x11\n\tmax_steps\x18\x05 \x01(\r\x12\x16\n\x0emax_inactivity\x18\x06 \x01(\r\x12\x19\n\x11nb_buffered_ticks\x18\x07 \x01(\x12\"/\n\nTrialActor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x61\x63tor_class\x18\x02 \x01(\t\"B\n\x0bObservation\x12\x0f\n\x07tick_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x06\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\x0c\"=\n\x06\x41\x63tion\x12\x0f\n\x07tick_id\x18\x01 \x01(\x12\x12\x11\n\ttimestamp\x18\x02 \x01(\x06\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\x0c\"o\n\x0cRewardSource\x12\x13\n\x0bsender_name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02\x12\x12\n\nconfidence\x18\x03 \x01(\x02\x12\'\n\tuser_data\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\"j\n\x06Reward\x12\x0f\n\x07tick_id\x18\x01 \x01(\x12\x12\x15\n\rreceiver_name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x02\x12)\n\x07sources\x18\x04 \x03(\x0b\x32\x18.cogmentAPI.RewardSource\"m\n\x07Message\x12\x0f\n\x07tick_id\x18\x01 \x01(\x12\x12\x13\n\x0bsender_name\x18\x02 \x01(\t\x12\x15\n\rreceiver_name\x18\x03 \x01(\t\x12%\n\x07payload\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\"\x91\x02\n\x12\x41\x63torRunTrialInput\x12-\n\x05state\x18\x01 \x01(\x0e\x32\x1e.cogmentAPI.CommunicationState\x12\x33\n\ninit_input\x18\x02 \x01(\x0b\x32\x1d.cogmentAPI.ActorInitialInputH\x00\x12.\n\x0bobservation\x18\x03 \x01(\x0b\x32\x17.cogmentAPI.ObservationH\x00\x12$\n\x06reward\x18\x04 \x01(\x0b\x32\x12.cogmentAPI.RewardH\x00\x12&\n\x07message\x18\x05 \x01(\x0b\x32\x13.cogmentAPI.MessageH\x00\x12\x11\n\x07\x64\x65tails\x18\x06 \x01(\tH\x00\x42\x06\n\x04\x64\x61ta\"\x8a\x02\n\x13\x41\x63torRunTrialOutput\x12-\n\x05state\x18\x01 \x01(\x0e\x32\x1e.cogmentAPI.CommunicationState\x12\x35\n\x0binit_output\x18\x02 \x01(\x0b\x32\x1e.cogmentAPI.ActorInitialOutputH\x00\x12$\n\x06\x61\x63tion\x18\x03 \x01(\x0b\x32\x12.cogmentAPI.ActionH\x00\x12$\n\x06reward\x18\x04 \x01(\x0b\x32\x12.cogmentAPI.RewardH\x00\x12&\n\x07message\x18\x05 \x01(\x0b\x32\x13.cogmentAPI.MessageH\x00\x12\x11\n\x07\x64\x65tails\x18\x06 \x01(\tH\x00\x42\x06\n\x04\x64\x61ta\"\x90\x01\n\x11\x41\x63torInitialInput\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x13\n\x0b\x61\x63tor_class\x18\x02 \x01(\t\x12\x11\n\timpl_name\x18\x03 \x01(\t\x12\x10\n\x08\x65nv_name\x18\x04 \x01(\t\x12-\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\"S\n\x12\x41\x63torInitialOutput\x12\x15\n\x0b\x61\x63tor_class\x18\x01 \x01(\tH\x00\x12\x14\n\nactor_name\x18\x02 \x01(\tH\x00\x42\x10\n\x0eslot_selection\"\x1e\n\x0bTrialConfig\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"\x1e\n\x0b\x41\x63torConfig\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"$\n\x11\x45nvironmentConfig\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c*a\n\nTrialState\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x10\n\x0cINITIALIZING\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x0b\n\x07RUNNING\x10\x03\x12\x0f\n\x0bTERMINATING\x10\x04\x12\t\n\x05\x45NDED\x10\x05*g\n\x12\x43ommunicationState\x12\x15\n\x11UNKNOWN_COM_STATE\x10\x00\x12\n\n\x06NORMAL\x10\x01\x12\r\n\tHEARTBEAT\x10\x02\x12\x08\n\x04LAST\x10\x03\x12\x0c\n\x08LAST_ACK\x10\x04\x12\x07\n\x03\x45ND\x10\x05\x42\x30Z.github.com/cogment/cogment/grpcapi/cogment/apib\x06proto3')

_TRIALSTATE = DESCRIPTOR.enum_types_by_name['TrialState']
TrialState = enum_type_wrapper.EnumTypeWrapper(_TRIALSTATE)
_COMMUNICATIONSTATE = DESCRIPTOR.enum_types_by_name['CommunicationState']
CommunicationState = enum_type_wrapper.EnumTypeWrapper(_COMMUNICATIONSTATE)
UNKNOWN = 0
INITIALIZING = 1
PENDING = 2
RUNNING = 3
TERMINATING = 4
ENDED = 5
UNKNOWN_COM_STATE = 0
NORMAL = 1
HEARTBEAT = 2
LAST = 3
LAST_ACK = 4
END = 5


_VERSIONREQUEST = DESCRIPTOR.message_types_by_name['VersionRequest']
_VERSIONINFO = DESCRIPTOR.message_types_by_name['VersionInfo']
_VERSIONINFO_VERSION = _VERSIONINFO.nested_types_by_name['Version']
_SERIALIZEDMESSAGE = DESCRIPTOR.message_types_by_name['SerializedMessage']
_DATALOGPARAMS = DESCRIPTOR.message_types_by_name['DatalogParams']
_ENVIRONMENTPARAMS = DESCRIPTOR.message_types_by_name['EnvironmentParams']
_ACTORPARAMS = DESCRIPTOR.message_types_by_name['ActorParams']
_TRIALPARAMS = DESCRIPTOR.message_types_by_name['TrialParams']
_TRIALACTOR = DESCRIPTOR.message_types_by_name['TrialActor']
_OBSERVATION = DESCRIPTOR.message_types_by_name['Observation']
_ACTION = DESCRIPTOR.message_types_by_name['Action']
_REWARDSOURCE = DESCRIPTOR.message_types_by_name['RewardSource']
_REWARD = DESCRIPTOR.message_types_by_name['Reward']
_MESSAGE = DESCRIPTOR.message_types_by_name['Message']
_ACTORRUNTRIALINPUT = DESCRIPTOR.message_types_by_name['ActorRunTrialInput']
_ACTORRUNTRIALOUTPUT = DESCRIPTOR.message_types_by_name['ActorRunTrialOutput']
_ACTORINITIALINPUT = DESCRIPTOR.message_types_by_name['ActorInitialInput']
_ACTORINITIALOUTPUT = DESCRIPTOR.message_types_by_name['ActorInitialOutput']
_TRIALCONFIG = DESCRIPTOR.message_types_by_name['TrialConfig']
_ACTORCONFIG = DESCRIPTOR.message_types_by_name['ActorConfig']
_ENVIRONMENTCONFIG = DESCRIPTOR.message_types_by_name['EnvironmentConfig']
VersionRequest = _reflection.GeneratedProtocolMessageType('VersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _VERSIONREQUEST,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.VersionRequest)
  })
_sym_db.RegisterMessage(VersionRequest)

VersionInfo = _reflection.GeneratedProtocolMessageType('VersionInfo', (_message.Message,), {

  'Version' : _reflection.GeneratedProtocolMessageType('Version', (_message.Message,), {
    'DESCRIPTOR' : _VERSIONINFO_VERSION,
    '__module__' : 'cogment.api.common_pb2'
    # @@protoc_insertion_point(class_scope:cogmentAPI.VersionInfo.Version)
    })
  ,
  'DESCRIPTOR' : _VERSIONINFO,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.VersionInfo)
  })
_sym_db.RegisterMessage(VersionInfo)
_sym_db.RegisterMessage(VersionInfo.Version)

SerializedMessage = _reflection.GeneratedProtocolMessageType('SerializedMessage', (_message.Message,), {
  'DESCRIPTOR' : _SERIALIZEDMESSAGE,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.SerializedMessage)
  })
_sym_db.RegisterMessage(SerializedMessage)

DatalogParams = _reflection.GeneratedProtocolMessageType('DatalogParams', (_message.Message,), {
  'DESCRIPTOR' : _DATALOGPARAMS,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.DatalogParams)
  })
_sym_db.RegisterMessage(DatalogParams)

EnvironmentParams = _reflection.GeneratedProtocolMessageType('EnvironmentParams', (_message.Message,), {
  'DESCRIPTOR' : _ENVIRONMENTPARAMS,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvironmentParams)
  })
_sym_db.RegisterMessage(EnvironmentParams)

ActorParams = _reflection.GeneratedProtocolMessageType('ActorParams', (_message.Message,), {
  'DESCRIPTOR' : _ACTORPARAMS,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorParams)
  })
_sym_db.RegisterMessage(ActorParams)

TrialParams = _reflection.GeneratedProtocolMessageType('TrialParams', (_message.Message,), {
  'DESCRIPTOR' : _TRIALPARAMS,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialParams)
  })
_sym_db.RegisterMessage(TrialParams)

TrialActor = _reflection.GeneratedProtocolMessageType('TrialActor', (_message.Message,), {
  'DESCRIPTOR' : _TRIALACTOR,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialActor)
  })
_sym_db.RegisterMessage(TrialActor)

Observation = _reflection.GeneratedProtocolMessageType('Observation', (_message.Message,), {
  'DESCRIPTOR' : _OBSERVATION,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.Observation)
  })
_sym_db.RegisterMessage(Observation)

Action = _reflection.GeneratedProtocolMessageType('Action', (_message.Message,), {
  'DESCRIPTOR' : _ACTION,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.Action)
  })
_sym_db.RegisterMessage(Action)

RewardSource = _reflection.GeneratedProtocolMessageType('RewardSource', (_message.Message,), {
  'DESCRIPTOR' : _REWARDSOURCE,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.RewardSource)
  })
_sym_db.RegisterMessage(RewardSource)

Reward = _reflection.GeneratedProtocolMessageType('Reward', (_message.Message,), {
  'DESCRIPTOR' : _REWARD,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.Reward)
  })
_sym_db.RegisterMessage(Reward)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.Message)
  })
_sym_db.RegisterMessage(Message)

ActorRunTrialInput = _reflection.GeneratedProtocolMessageType('ActorRunTrialInput', (_message.Message,), {
  'DESCRIPTOR' : _ACTORRUNTRIALINPUT,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorRunTrialInput)
  })
_sym_db.RegisterMessage(ActorRunTrialInput)

ActorRunTrialOutput = _reflection.GeneratedProtocolMessageType('ActorRunTrialOutput', (_message.Message,), {
  'DESCRIPTOR' : _ACTORRUNTRIALOUTPUT,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorRunTrialOutput)
  })
_sym_db.RegisterMessage(ActorRunTrialOutput)

ActorInitialInput = _reflection.GeneratedProtocolMessageType('ActorInitialInput', (_message.Message,), {
  'DESCRIPTOR' : _ACTORINITIALINPUT,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorInitialInput)
  })
_sym_db.RegisterMessage(ActorInitialInput)

ActorInitialOutput = _reflection.GeneratedProtocolMessageType('ActorInitialOutput', (_message.Message,), {
  'DESCRIPTOR' : _ACTORINITIALOUTPUT,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorInitialOutput)
  })
_sym_db.RegisterMessage(ActorInitialOutput)

TrialConfig = _reflection.GeneratedProtocolMessageType('TrialConfig', (_message.Message,), {
  'DESCRIPTOR' : _TRIALCONFIG,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialConfig)
  })
_sym_db.RegisterMessage(TrialConfig)

ActorConfig = _reflection.GeneratedProtocolMessageType('ActorConfig', (_message.Message,), {
  'DESCRIPTOR' : _ACTORCONFIG,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActorConfig)
  })
_sym_db.RegisterMessage(ActorConfig)

EnvironmentConfig = _reflection.GeneratedProtocolMessageType('EnvironmentConfig', (_message.Message,), {
  'DESCRIPTOR' : _ENVIRONMENTCONFIG,
  '__module__' : 'cogment.api.common_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvironmentConfig)
  })
_sym_db.RegisterMessage(EnvironmentConfig)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z.github.com/cogment/cogment/grpcapi/cogment/api'
  _TRIALSTATE._serialized_start=2356
  _TRIALSTATE._serialized_end=2453
  _COMMUNICATIONSTATE._serialized_start=2455
  _COMMUNICATIONSTATE._serialized_end=2558
  _VERSIONREQUEST._serialized_start=67
  _VERSIONREQUEST._serialized_end=83
  _VERSIONINFO._serialized_start=85
  _VERSIONINFO._serialized_end=191
  _VERSIONINFO_VERSION._serialized_start=151
  _VERSIONINFO_VERSION._serialized_end=191
  _SERIALIZEDMESSAGE._serialized_start=193
  _SERIALIZEDMESSAGE._serialized_end=229
  _DATALOGPARAMS._serialized_start=231
  _DATALOGPARAMS._serialized_end=288
  _ENVIRONMENTPARAMS._serialized_start=290
  _ENVIRONMENTPARAMS._serialized_end=412
  _ACTORPARAMS._serialized_start=415
  _ACTORPARAMS._serialized_end=687
  _TRIALPARAMS._serialized_start=690
  _TRIALPARAMS._serialized_end=963
  _TRIALACTOR._serialized_start=965
  _TRIALACTOR._serialized_end=1012
  _OBSERVATION._serialized_start=1014
  _OBSERVATION._serialized_end=1080
  _ACTION._serialized_start=1082
  _ACTION._serialized_end=1143
  _REWARDSOURCE._serialized_start=1145
  _REWARDSOURCE._serialized_end=1256
  _REWARD._serialized_start=1258
  _REWARD._serialized_end=1364
  _MESSAGE._serialized_start=1366
  _MESSAGE._serialized_end=1475
  _ACTORRUNTRIALINPUT._serialized_start=1478
  _ACTORRUNTRIALINPUT._serialized_end=1751
  _ACTORRUNTRIALOUTPUT._serialized_start=1754
  _ACTORRUNTRIALOUTPUT._serialized_end=2020
  _ACTORINITIALINPUT._serialized_start=2023
  _ACTORINITIALINPUT._serialized_end=2167
  _ACTORINITIALOUTPUT._serialized_start=2169
  _ACTORINITIALOUTPUT._serialized_end=2252
  _TRIALCONFIG._serialized_start=2254
  _TRIALCONFIG._serialized_end=2284
  _ACTORCONFIG._serialized_start=2286
  _ACTORCONFIG._serialized_end=2316
  _ENVIRONMENTCONFIG._serialized_start=2318
  _ENVIRONMENTCONFIG._serialized_end=2354
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.common_pb2 as common_api

from cogment.session import Session
from cogment.errors import CogmentError
from cogment.utils import logger

import time


class ActorClass:
    """Class containing the details of an actor class defined in a config file."""

    def __init__(self, name, config_type, action_space, observation_space):
        self.name = name
        self.config_type = config_type
        self.action_space = action_space
        self.observation_space = observation_space

    def __str__(self):
        result = f"ActorClass: name = {self.name}, config_type = {type(self.config_type)}"
        result += f", action_space = {type(self.action_space)}, observation_space = {type(self.observation_space)}"
        return result


class ActorClassList:
    """Class containing the list of actor classes defined in a config file."""

    def __init__(self, *args):
        self._actor_classes_list = list(args)

        for a_c in args:
            setattr(self, a_c.name, a_c)

    def __iter__(self):
        return iter(self._actor_classes_list)

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def get_class_by_index(self, index):
        return self._actor_classes_list[index]

    def __str__(self):
        result = f"ActorClassList:"
        for ac in self._actor_classes_list:
            result += f" {ac.name} = {ac},"
        return result

    def get(self, key, default=None):  # Similar to dict.get()
        return getattr(self, key, default)


class ActorSession(Session):
    """Derived class representing the session of an actor for a trial."""

    def __init__(self, impl, actor_class, trial, name, impl_name, env_name, config):
        super().__init__(trial, name, impl, impl_name, config)
        self.class_name = actor_class.name
        self.env_name = env_name
        self._actor_class = actor_class

    def __str__(self):
        result = super().__str__()
        result += f" --- ActorSession: class_name = {self.class_name}, config = {self.config}"
        return result

    def get_active_actors(self):
        # Controller.get_actors can be used.
        # Or provide actors details in the config, or the observations.
        raise CogmentError(f"This function is deprecated for actors")

    def start(self, auto_done_sending=True):
        self._start(auto_done_sending)

    def do_action(self, action):
        action_req = common_api.Action()
        action_req.timestamp = int(time.time() * 1e9)
        action_req.tick_id = self._last_tick_delivered
        if action is not None:
            action_req.content = action.SerializeToString()

        self._post_outgoing_data(action_req)

    def send_message(self, payload, to, to_environment=None):
        if to_environment is not None:
            logger.deprecated("Parameter 'to_environment' is deprecated for 'send_message' method. "
                              "Use 'self.env_name' as environment name in the 'to' parameter.")
            if to_environment:
                self._send_message(payload, to + [self.env_name])
                return
        self._send_message(payload, to)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc

import cogment.api.common_pb2 as common_api

from cogment.version import __version__

import logging
import types
import os


def _logger_trace(self, msg, *args, **kwargs):
    """Simplified trace log output method"""
    self.log(self.TRACE, msg, *args, **kwargs)


def _logger_deprecated(self, msg):
    """Log deprecation as warning but only once"""
    if msg not in self.DEPRECATED_LOGGED:
        self.warning(msg)
        self.DEPRECATED_LOGGED.add(msg)


_TRACE = 5

logger = logging.getLogger("cogment.sdk")
setattr(logger, "TRACE", _TRACE)  # We don't want a new global level name (i.e. with 'logging.addLevelName')
setattr(logger, "trace", types.MethodType(_logger_trace, logger))
setattr(logger, "DEPRECATED_LOGGED", set())
setattr(logger, "deprecated", types.MethodType(_logger_deprecated, logger))

_COGMENT_LOG_LEVEL = os.environ.get("COGMENT_LOG_LEVEL", "INFO").upper()
if _COGMENT_LOG_LEVEL == "OFF":
    logger.setLevel(logging.CRITICAL)
elif _COGMENT_LOG_LEVEL == "TRACE":
    logger.setLevel(_TRACE)  # stupid pylint can't realize that 'TRACE' is part of 'logger'!
else:
    logger.setLevel(_COGMENT_LOG_LEVEL)
logger.addHandler(logging.NullHandler())


# Timeout (in seconds) for the init data to come from the Orchestrator before we consider a failure.
INIT_TIMEOUT = 30


def list_versions():
    reply = common_api.VersionInfo()
    reply.versions.add(name='cogment_sdk', version=__version__)
    reply.versions.add(name='grpc', version=grpc.__version__)

    return reply



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.errors import CogmentError
from cogment.utils import logger


class PrehookSession():
    """Class representing the session of a pre-trial hook for a trial."""

    def __init__(self, trial, user_id, trial_params):
        self._trial = trial
        self._user_id = user_id
        self._data_attributes = {"trial_config", "trial_max_steps", "trial_max_inactivity",
                      "datalog_endpoint", "datalog_exclude", "environment_implementation",
                      "environment_endpoint", "environment_name", "environment_config", "actors"}
        self._new_attributes = {"trial_parameters"}
        self._method_attributes = {"get_trial_id", "get_user_id", "validate"}
        self._attributes = self._data_attributes.union(self._method_attributes).union(self._new_attributes)
        self._actor_attributes = {"name", "actor_class", "endpoint", "implementation", "config"}

        self._trial_parameters = trial_params
        self._trial_parameters_use = False

    def __str__(self):
        result = "PreHookSession:"
        for att in self._data_attributes :
            if hasattr(self, att):
                result += f", {att} = {getattr(self, att)}"
        result += f" trial_parameters = {self._trial_parameters}"

        return result

    def get_trial_id(self):
        return self._trial.id

    def get_user_id(self):
        return self._user_id

    def validate(self):
        if self._trial_parameters_use:
            for att in dir(self):
                if att in self._data_attributes:
                    raise CogmentError(f"Cannot mix deprecated attribute [{att}] with new attribute 'trial_parameters'")
        else:
            for att in dir(self):
                if att and att[0] != "_" and att not in self._attributes:
                    raise CogmentError(f"Unknown attribute [{att}] for parameters")
            for att in self._attributes:
                if att not in dir(self):
                    raise CogmentError(f"Missing attribute [{att}] for parameters")

            for actor in self.actors:
                for att in self._actor_attributes:
                    if att not in actor:
                        raise CogmentError(f"Incomplete actor for parameters. Missing attribute [{att}]")
                for att in actor:
                    if att not in self._actor_attributes:
                        raise CogmentError(f"Unknown actor attribute [{att}] for parameters")

    @property
    def trial_parameters(self):
        if not self._trial_parameters_use:
            self._trial_parameters_use = True
            for att in self._data_attributes:
                if hasattr(self, att):
                    delattr(self, att)

        return self._trial_parameters

    @trial_parameters.setter
    def trial_parameters(self, val):
        if not self._trial_parameters_use:
            self._trial_parameters_use = True
            for att in self._data_attributes:
                if hasattr(self, att):
                    delattr(self, att)

        self._trial_parameters = val



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.aio  # type: ignore
from prometheus_client import start_http_server as start_prometheus_server
from prometheus_client.core import REGISTRY  # type: ignore

import cogment.api.orchestrator_pb2_grpc as orchestrator_grpc_api
import cogment.api.agent_pb2_grpc as agent_grpc_api
import cogment.api.environment_pb2_grpc as env_grpc_api
import cogment.api.hooks_pb2_grpc as hooks_grpc_api
import cogment.api.datalog_pb2_grpc as datalog_grpc_api
import cogment.api.trial_datastore_pb2_grpc as datastore_grpc_api
import cogment.api.model_registry_pb2_grpc as model_registry_api
import cogment.api.directory_pb2_grpc as directory_grpc_api

import cogment.endpoints as ep
from cogment.directory import Directory, ServiceType
from cogment.actor import ActorSession
from cogment.environment import EnvironmentSession
from cogment.prehook import PrehookSession
from cogment.datalog import DatalogSession
from cogment.datastore import Datastore
from cogment.model_registry import ModelRegistry
from cogment.control import Controller
from cogment.agent_service import AgentServicer, get_actor_impl
from cogment.client_service import ClientServicer
from cogment.env_service import EnvironmentServicer
from cogment.hooks_service import PrehookServicer
from cogment.datalog_service import DatalogServicer
from cogment.errors import CogmentError
from cogment.utils import logger
from cogment.version import __version__

import os
from typing import Callable, Awaitable, Dict, List, Any, Tuple
from types import ModuleType
import asyncio
from types import SimpleNamespace
import socket
import urllib.parse as urlpar


DEFAULT_PROMETHEUS_PORT = 8000
_ADDITIONAL_REGISTRATION_ITEMS = {"__registration_source" : "PythonSDK-Implicit", "__version" : __version__}


def _make_client_channel(grpc_endpoint: ep.Endpoint):
    parsed_url = urlpar.urlparse(grpc_endpoint.url)
    if parsed_url.scheme == ep.GRPC_SCHEME:
        url = parsed_url.netloc
    else:
        raise CogmentError(f"Invalid endpoint scheme (must be 'grpc') [{grpc_endpoint.url}]")

    if not grpc_endpoint.using_ssl():
        channel = grpc.aio.insecure_channel(url)
    else:
        if grpc_endpoint.root_certificates:
            root = bytes(grpc_endpoint.root_certificates, "utf-8")
        else:
            root = None
        if grpc_endpoint.private_key:
            key = bytes(grpc_endpoint.private_key, "utf-8")
        else:
            key = None
        if grpc_endpoint.certificate_chain:
            certs = bytes(grpc_endpoint.certificate_chain, "utf-8")
        else:
            certs = None
        creds = grpc.ssl_channel_credentials(root, key, certs)
        channel = grpc.aio.secure_channel(url, creds)

    return channel


def _make_server(served_endpoint: ep.ServedEndpoint):
    server = grpc.aio.server()

    address = f"[::]:{served_endpoint.port}"
    if not served_endpoint.using_ssl():
        server.add_insecure_port(address)
    else:
        if served_endpoint.root_certificates:
            require_client_auth = True
            root = bytes(served_endpoint.root_certificates, "utf-8")
        else:
            require_client_auth = False
            root = None
        certs = []
        for (key, chain) in served_endpoint.private_key_certificate_chain_pairs:
            certs.append((bytes(key, "utf-8"), bytes(chain, "utf-8")))
        if not certs:
            certs = None

        creds = grpc.ssl_server_credentials(certs, root, require_client_auth)
        server.add_secure_port(address, creds)

    return server


class Context:
    """Top level class for the Cogment library from which to obtain all services."""

    def __init__(self, user_id: str, cog_settings: ModuleType, asyncio_loop=None,
                       prometheus_registry=REGISTRY, directory_endpoint: ep.Endpoint = None,
                       directory_auth_token: str = None):
        self._user_id = user_id
        self._actor_impls: Dict[str, SimpleNamespace] = {}
        self._env_impls: Dict[str, SimpleNamespace] = {}
        self._prehook_impl: SimpleNamespace = None
        self._datalog_impl: SimpleNamespace = None
        self._grpc_server = None  # type: Any
        self._prometheus_registry = prometheus_registry
        self._cog_settings = cog_settings

        if asyncio_loop is None:
            # Make sure we are running in a asyncio.Task. Even if technically this init does not need
            # to be in a Task, the whole of the SDK expects to be running in Tasks of the same event loop.
            try:
                self.asyncio_loop = asyncio.get_running_loop()
            except RuntimeError:
                raise CogmentError("The Cogment SDK requires running in a Python `asyncio` Task")
        else:
            self.asyncio_loop = asyncio_loop

        if directory_endpoint is not None:
            endpoint_to_use = directory_endpoint
        else:
            env_endpoint = os.environ.get("COGMENT_DIRECTORY_ENDPOINT")
            if env_endpoint is not None:
                endpoint_to_use = ep.Endpoint(env_endpoint)
            else:
                endpoint_to_use = None

        if directory_auth_token is not None:
            auth_token_to_use = directory_auth_token
        else:
            auth_token_to_use = os.environ.get("COGMENT_DIRECTORY_AUTHENTICATION_TOKEN")

        if endpoint_to_use:
            try:
                channel = _make_client_channel(endpoint_to_use)
            except CogmentError as exc:
                raise CogmentError(f"Directory endpoint: {exc}")

            stub = directory_grpc_api.DirectorySPStub(channel)
            self._directory = Directory(stub, auth_token_to_use)
        else:
            self._directory = None

    def __str__(self):
        result = f"Cogment Context: user id = {self._user_id}"
        return result

    def register_actor(self,
                       impl: Callable[[ActorSession], Awaitable[None]],
                       impl_name: str,
                       actor_classes: List[str] = [], properties: Dict[str, str] = {}):

        if self._grpc_server is not None:
            # We could accept "client" actor registration after the server is started, but it is not worth it
            raise CogmentError("Cannot register an actor after the server is started")
        if impl_name in self._actor_impls:
            raise CogmentError(f"The actor implementation name must be unique: [{impl_name}]")
        if ep.ACTOR_CLASS_PROPERTY_NAME in properties:
            raise CogmentError(f"Actor property [{ep.ACTOR_CLASS_PROPERTY_NAME}] is reserved for internal use")
        if ep.IMPLEMENTATION_PROPERTY_NAME in properties:
            raise CogmentError(f"Actor property [{ep.IMPLEMENTATION_PROPERTY_NAME}] is reserved for internal use")

        directory_properties = {}
        directory_properties.update(properties)
        directory_properties[ep.IMPLEMENTATION_PROPERTY_NAME] = impl_name
        directory_properties.update(_ADDITIONAL_REGISTRATION_ITEMS)

        if len(actor_classes) > 0:
            directory_actor_classes = [ac for ac in actor_classes]
        else:
            logger.deprecated(f"The use of an empty list of actor_classes is deprecated")  # type: ignore
            directory_actor_classes = []
            for ac in self._cog_settings.actor_classes:
                actor_classes.append(ac.name)
            logger.warning(f"Implementation [{impl_name}] will be registered to the directory"
                            f" for all actor classes [{actor_classes}]")

        self._actor_impls[impl_name] = SimpleNamespace(
            impl=impl, actor_classes=directory_actor_classes, properties=directory_properties)

    def register_environment(self,
                             impl: Callable[[EnvironmentSession], Awaitable[None]],
                             impl_name: str = "default", properties: Dict[str, str] = {}):
        if self._grpc_server is not None:
            raise CogmentError("Cannot register an environment after the server is started")
        if impl_name in self._env_impls:
            raise CogmentError(f"The environment implementation name must be unique: [{impl_name}]")
        if ep.IMPLEMENTATION_PROPERTY_NAME in properties:
            raise CogmentError(f"Environment property [{ep.IMPLEMENTATION_PROPERTY_NAME}] is reserved for internal use")

        directory_properties = {}
        directory_properties.update(properties)
        directory_properties[ep.IMPLEMENTATION_PROPERTY_NAME] = impl_name
        directory_properties.update(_ADDITIONAL_REGISTRATION_ITEMS)

        self._env_impls[impl_name] = SimpleNamespace(impl=impl, properties=directory_properties)

    def register_pre_trial_hook(self,
                                impl: Callable[[PrehookSession], Awaitable[None]],
                                properties: Dict[str, str] = {}):
        if self._grpc_server is not None:
            raise CogmentError("Cannot register a pre-trial hook after the server is started")
        if self._prehook_impl is not None:
            raise CogmentError("Only one pre-trial hook service can be registered")

        directory_properties = {}
        directory_properties.update(properties)
        directory_properties.update(_ADDITIONAL_REGISTRATION_ITEMS)
        self._prehook_impl = SimpleNamespace(impl=impl, properties=directory_properties)

    def register_datalog(self,
                         impl: Callable[[DatalogSession], Awaitable[None]],
                         properties: Dict[str, str] = {}):
        if self._grpc_server is not None:
            raise CogmentError("Cannot register a datalog after the server is started")
        if self._datalog_impl is not None:
            raise CogmentError("Only one datalog service can be registered")

        directory_properties = {}
        directory_properties.update(properties)
        directory_properties.update(_ADDITIONAL_REGISTRATION_ITEMS)
        self._datalog_impl = SimpleNamespace(impl=impl, properties=directory_properties)

    async def _directory_deregistration(self, registered: List[Tuple[int, str]]):
        if self._directory is None:
            return

        for item in registered:
            try:
                await self._directory.deregister_service(*item)
            except Exception as exc:
                logger.debug(f"Failed to deregister service id [{item[0]}] from directory: [{exc}]")

        registered.clear()

    async def _directory_registration(self, port, ssl):
        registered: List[Tuple[int, str]] = []

        if self._directory is None:
            logger.debug(f"No directory for implicit service registration")
            return registered

        socket_addr = socket.gethostbyname(socket.gethostname())
        logger.debug(f"Registering services available here [{socket_addr}] to Directory")

        try:
            for actor in self._actor_impls.values():
                properties = actor.properties
                for actor_class in actor.actor_classes:
                    properties[ep.ACTOR_CLASS_PROPERTY_NAME] = actor_class
                    res = await self._directory.register_host(
                        ServiceType.ACTOR, socket_addr, port, ssl, properties)
                    registered.append(res)

            for env in self._env_impls.values():
                res = await self._directory.register_host(
                    ServiceType.ENVIRONMENT, socket_addr, port, ssl, env.properties)
                registered.append(res)

            if self._prehook_impl is not None:
                res = await self._directory.register_host(
                    ServiceType.HOOK, socket_addr, port, ssl, self._prehook_impl.properties)
                registered.append(res)

            if self._datalog_impl is not None:
                res = await self._directory.register_host(
                    ServiceType.DATALOG, socket_addr, port, ssl, self._datalog_impl.properties)
                registered.append(res)

        except Exception:
            await self._directory_deregistration(registered)
            raise

        return registered

    async def serve_all_registered(self, served_endpoint: ep.ServedEndpoint, prometheus_port=DEFAULT_PROMETHEUS_PORT):
        if (len(self._actor_impls) == 0 and len(self._env_impls) == 0 and
                self._prehook_impl is None and self._datalog_impl is None):
            raise CogmentError("Nothing registered to serve!")
        if self._grpc_server is not None:
            raise CogmentError("Cannot serve the same components twice")

        self._grpc_server = _make_server(served_endpoint)

        if self._actor_impls:
            servicer = AgentServicer(self._actor_impls, self._cog_settings, self._prometheus_registry)
            agent_grpc_api.add_ServiceActorSPServicer_to_server(servicer, self._grpc_server)

        if self._env_impls:
            servicer = EnvironmentServicer(self._env_impls, self._cog_settings, self._prometheus_registry)
            env_grpc_api.add_EnvironmentSPServicer_to_server(servicer, self._grpc_server)

        if self._prehook_impl is not None:
            servicer = PrehookServicer(self._prehook_impl.impl, self._cog_settings, self._prometheus_registry)
            hooks_grpc_api.add_TrialHooksSPServicer_to_server(servicer, self._grpc_server)

        if self._datalog_impl is not None:
            servicer = DatalogServicer(self._datalog_impl.impl, self._cog_settings)
            datalog_grpc_api.add_DatalogSPServicer_to_server(servicer, self._grpc_server)

        if self._prometheus_registry is not None and prometheus_port is not None:
            start_prometheus_server(prometheus_port, "", self._prometheus_registry)

        await self._grpc_server.start()
        logger.debug(f"Context gRPC server at port [{served_endpoint.port}] for user [{self._user_id}] started")

        directory_registered = await self._directory_registration(served_endpoint.port, served_endpoint.using_ssl())

        try:
            await self._grpc_server.wait_for_termination()
            logger.debug(f"Context gRPC server at port [{served_endpoint.port}] for user [{self._user_id}] exited")

        finally:
            await self._directory_deregistration(directory_registered)

    def _make_controller(self, endpoint):
        channel = _make_client_channel(endpoint)
        stub = orchestrator_grpc_api.TrialLifecycleSPStub(channel)
        return Controller(stub, self._user_id)

    async def _inquire_and_make_controller(self, endpoint):
        inquired_endpoint = await self._directory.get_inquired_endpoint(endpoint, ServiceType.LIFE_CYCLE)
        return self._make_controller(inquired_endpoint)

    # TODO: The non-async part is only kept for backward compatibility,
    #       remove it in a future (backward incompatible) release.
    def get_controller(self, endpoint=ep.Endpoint()):
        try:
            parsed_url = urlpar.urlparse(endpoint.url)
        except Exception as exc:
            raise CogmentError(f"Endpoint [{endpoint.url}]: {exc}")

        if parsed_url.scheme == ep.GRPC_SCHEME or self._directory is None:
            return self._make_controller(endpoint)  # This returns a controller instance
        else:
            return self._inquire_and_make_controller(endpoint)  # This returns an awaitable object

    def _make_datastore(self, endpoint):
        channel = _make_client_channel(endpoint)
        stub = datastore_grpc_api.TrialDatastoreSPStub(channel)
        return Datastore(stub, self._cog_settings)

    async def _inquire_and_make_datastore(self, endpoint):
        inquired_endpoint = await self._directory.get_inquired_endpoint(endpoint, ServiceType.DATASTORE)
        return self._make_datastore(inquired_endpoint)

    # TODO: The non-async part is only kept for backward compatibility,
    #       remove it in a future (backward incompatible) release.
    def get_datastore(self, endpoint=ep.Endpoint()):
        try:
            parsed_url = urlpar.urlparse(endpoint.url)
        except Exception as exc:
            raise CogmentError(f"Endpoint [{endpoint.url}]: {exc}")

        if parsed_url.scheme == ep.GRPC_SCHEME or self._directory is None:
            return self._make_datastore(endpoint)  # This returns a datastore instance
        else:
            return self._inquire_and_make_datastore(endpoint)  # This returns an awaitable object

    # Undocumented
    # We may want to make it async to standardize with the future
    # versions of 'get_controller' and 'get_datastore'
    def get_directory(self, endpoint: ep.Endpoint, authentication_token: str = None):
        channel = _make_client_channel(endpoint)
        stub = directory_grpc_api.DirectorySPStub(channel)
        return Directory(stub, authentication_token)

    async def get_model_registry(self, endpoint=ep.Endpoint()):
        if self._directory is not None:
            endpoint = await self._directory.get_inquired_endpoint(endpoint, ServiceType.MODEL_REG)

        channel = _make_client_channel(endpoint)
        stub = model_registry_api.ModelRegistrySPStub(channel)
        return ModelRegistry(stub)

    async def join_trial(self, trial_id, endpoint=ep.Endpoint(), impl_name=None, actor_name=None, actor_class=None):
        requested_class = None
        requested_name = None
        if impl_name is not None:
            # For backward compatibility
            logger.deprecated(f"`join_trial` parameter `impl_name` is deprecated")
            if actor_name is None:
                actor_impl = self._actor_impls[impl_name]
                if len(actor_impl.actor_classes) == 0:
                    raise CogmentError(f"Unable to determine possible actor to join trial: "
                                       f"impl_name [{impl_name}] does not have any registered actor class")
                requested_class = actor_impl.actor_classes[0]
            else:
                requested_name = actor_name

        elif actor_name is not None:
            requested_name = actor_name
            if actor_class is not None:
                logger.warning(f"`actor_class` will be ignored because `actor_name` is provided in `join_trial")

        elif actor_class is not None:
            requested_class = actor_class

        else:
            raise CogmentError(f"Actor name or actor class must be specified to join a trial")

        if self._directory is not None:
            inquired_endpoint = await self._directory.get_inquired_endpoint(endpoint, ServiceType.CLIENT_ACTOR)
        else:
            inquired_endpoint = endpoint

        channel = _make_client_channel(inquired_endpoint)
        stub = orchestrator_grpc_api.ClientActorSPStub(channel)
        servicer = ClientServicer(self._cog_settings, stub)

        init_data = await servicer.join_trial(trial_id, requested_name, requested_class)
        if not init_data:
            return

        if requested_name is not None and requested_name != init_data.actor_name:
            raise CogmentError(f"Internal failure: Actor name [{requested_name}] requested, received: {init_data}")
        if requested_class is not None and requested_class != init_data.actor_class:
            raise CogmentError(f"Internal failure: Actor class [{requested_class}] requested, received: {init_data}")

        if impl_name is not None and init_data.impl_name and init_data.impl_name != impl_name:
            logger.warning(f"Requested impl_name [{impl_name}] does not match trial impl_name "
                           f"[{init_data.impl_name}]: Requested impl_name will be used.")
            init_data.impl_name = impl_name

        actor_impl = get_actor_impl(trial_id, self._actor_impls, init_data)

        await servicer.run_session(actor_impl.impl, init_data)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/environment.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1d\x63ogment/api/environment.proto\x12\ncogmentAPI\x1a\x18\x63ogment/api/common.proto\"^\n\x0eObservationSet\x12\x0f\n\x07tick_id\x18\x01 \x01(\x12\x12\x11\n\ttimestamp\x18\x02 \x01(\x06\x12\x14\n\x0cobservations\x18\x03 \x03(\x0c\x12\x12\n\nactors_map\x18\x04 \x03(\x05\"\\\n\tActionSet\x12\x0f\n\x07tick_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x06\x12\x0f\n\x07\x61\x63tions\x18\x03 \x03(\x0c\x12\x1a\n\x12unavailable_actors\x18\x04 \x03(\r\"\xe4\x01\n\x10\x45nvRunTrialInput\x12-\n\x05state\x18\x01 \x01(\x0e\x32\x1e.cogmentAPI.CommunicationState\x12\x31\n\ninit_input\x18\x02 \x01(\x0b\x32\x1b.cogmentAPI.EnvInitialInputH\x00\x12+\n\naction_set\x18\x03 \x01(\x0b\x32\x15.cogmentAPI.ActionSetH\x00\x12&\n\x07message\x18\x04 \x01(\x0b\x32\x13.cogmentAPI.MessageH\x00\x12\x11\n\x07\x64\x65tails\x18\x05 \x01(\tH\x00\x42\x06\n\x04\x64\x61ta\"\x97\x02\n\x11\x45nvRunTrialOutput\x12-\n\x05state\x18\x01 \x01(\x0e\x32\x1e.cogmentAPI.CommunicationState\x12\x33\n\x0binit_output\x18\x02 \x01(\x0b\x32\x1c.cogmentAPI.EnvInitialOutputH\x00\x12\x35\n\x0fobservation_set\x18\x03 \x01(\x0b\x32\x1a.cogmentAPI.ObservationSetH\x00\x12$\n\x06reward\x18\x04 \x01(\x0b\x32\x12.cogmentAPI.RewardH\x00\x12&\n\x07message\x18\x05 \x01(\x0b\x32\x13.cogmentAPI.MessageH\x00\x12\x11\n\x07\x64\x65tails\x18\x06 \x01(\tH\x00\x42\x06\n\x04\x64\x61ta\"\xa3\x01\n\x0f\x45nvInitialInput\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\timpl_name\x18\x02 \x01(\t\x12\x0f\n\x07tick_id\x18\x03 \x01(\x04\x12/\n\x0f\x61\x63tors_in_trial\x18\x04 \x03(\x0b\x32\x16.cogmentAPI.TrialActor\x12-\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessage\"\x12\n\x10\x45nvInitialOutput2\xa0\x01\n\rEnvironmentSP\x12M\n\x08RunTrial\x12\x1c.cogmentAPI.EnvRunTrialInput\x1a\x1d.cogmentAPI.EnvRunTrialOutput\"\x00(\x01\x30\x01\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x42\x30Z.github.com/cogment/cogment/grpcapi/cogment/apib\x06proto3')



_OBSERVATIONSET = DESCRIPTOR.message_types_by_name['ObservationSet']
_ACTIONSET = DESCRIPTOR.message_types_by_name['ActionSet']
_ENVRUNTRIALINPUT = DESCRIPTOR.message_types_by_name['EnvRunTrialInput']
_ENVRUNTRIALOUTPUT = DESCRIPTOR.message_types_by_name['EnvRunTrialOutput']
_ENVINITIALINPUT = DESCRIPTOR.message_types_by_name['EnvInitialInput']
_ENVINITIALOUTPUT = DESCRIPTOR.message_types_by_name['EnvInitialOutput']
ObservationSet = _reflection.GeneratedProtocolMessageType('ObservationSet', (_message.Message,), {
  'DESCRIPTOR' : _OBSERVATIONSET,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ObservationSet)
  })
_sym_db.RegisterMessage(ObservationSet)

ActionSet = _reflection.GeneratedProtocolMessageType('ActionSet', (_message.Message,), {
  'DESCRIPTOR' : _ACTIONSET,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ActionSet)
  })
_sym_db.RegisterMessage(ActionSet)

EnvRunTrialInput = _reflection.GeneratedProtocolMessageType('EnvRunTrialInput', (_message.Message,), {
  'DESCRIPTOR' : _ENVRUNTRIALINPUT,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvRunTrialInput)
  })
_sym_db.RegisterMessage(EnvRunTrialInput)

EnvRunTrialOutput = _reflection.GeneratedProtocolMessageType('EnvRunTrialOutput', (_message.Message,), {
  'DESCRIPTOR' : _ENVRUNTRIALOUTPUT,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvRunTrialOutput)
  })
_sym_db.RegisterMessage(EnvRunTrialOutput)

EnvInitialInput = _reflection.GeneratedProtocolMessageType('EnvInitialInput', (_message.Message,), {
  'DESCRIPTOR' : _ENVINITIALINPUT,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvInitialInput)
  })
_sym_db.RegisterMessage(EnvInitialInput)

EnvInitialOutput = _reflection.GeneratedProtocolMessageType('EnvInitialOutput', (_message.Message,), {
  'DESCRIPTOR' : _ENVINITIALOUTPUT,
  '__module__' : 'cogment.api.environment_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.EnvInitialOutput)
  })
_sym_db.RegisterMessage(EnvInitialOutput)

_ENVIRONMENTSP = DESCRIPTOR.services_by_name['EnvironmentSP']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z.github.com/cogment/cogment/grpcapi/cogment/api'
  _OBSERVATIONSET._serialized_start=71
  _OBSERVATIONSET._serialized_end=165
  _ACTIONSET._serialized_start=167
  _ACTIONSET._serialized_end=259
  _ENVRUNTRIALINPUT._serialized_start=262
  _ENVRUNTRIALINPUT._serialized_end=490
  _ENVRUNTRIALOUTPUT._serialized_start=493
  _ENVRUNTRIALOUTPUT._serialized_end=772
  _ENVINITIALINPUT._serialized_start=775
  _ENVINITIALINPUT._serialized_end=938
  _ENVINITIALOUTPUT._serialized_start=940
  _ENVINITIALOUTPUT._serialized_end=958
  _ENVIRONMENTSP._serialized_start=961
  _ENVIRONMENTSP._serialized_end=1121
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc.aio  # type: ignore

import cogment.api.datalog_pb2_grpc as datalog_grpc_api
import cogment.api.common_pb2 as common_api
import cogment.api.datalog_pb2 as datalog_api

from cogment.control import TrialState
from cogment.datalog import DatalogSession, LogParams
from cogment.parameters import TrialParameters
from cogment.errors import CogmentError
from cogment.utils import list_versions, logger
from cogment.session import RecvObservation, ActorStatus, RecvAction, RecvMessage, RecvReward

import asyncio


class LogSample:
    """Class representing a trial sample for the datalog service."""

    def __init__(self, params):
        self.tick_id = None
        self.out_of_sync = None
        self.timestamp = None
        self.state = None
        self.events = None

        self._raw_sample = None
        self._actor_indexes = None

        self._log_params = None
        self._parameters = None
        if type(params) == LogParams:
            logger.deprecated(f"Deprecated use of LogParams")
            self._log_params = params
            self._nb_actors = params.nb_actors
        elif type(params) == TrialParameters:
            self._parameters = params
            self._nb_actors = len(params.actors)
        else:
            raise CogmentError(f"Wrong type of params provided [{type(params)}]")

    def __str__(self):
        result = f"LogSample: {self._raw_sample}"
        return result

    # Type of serialized data being produced and consumed by this class.
    # This is dependent on all the underlying protobuf messages used to
    # serialize/deserialize, and should be incremented if any of them changes in
    # a backward or forward incompatible way. API 1.0 could be considered a type 0 or 1.
    # Current dependencies: DatalogSample, SampleInfo, TrialState, ObservationSet,
    #                       Action, Reward, RewardSource, Message
    def get_serialization_type(self):
        return 2

    def _set(self, raw_sample):
        self._set_from_sample(raw_sample)

    def serialize(self):
        if self._raw_sample is None:
            raise CogmentError("Not set, nothing to serialize")
        return self._raw_sample.SerializeToString()

    def deserialize(self, raw_string):
        sample = datalog_api.DatalogSample()
        sample.ParseFromString(raw_string)
        self._set_from_sample(sample)

    def _set_from_sample(self, sample):
        if type(sample) != datalog_api.DatalogSample:
            raise CogmentError(f"Wrong type of sample provided [{type(sample)}]")
        if sample.HasField("info"):
            self.state = TrialState(sample.info.state)
            self.out_of_sync = sample.info.out_of_sync
            self.tick_id = sample.info.tick_id
            self.timestamp = sample.info.timestamp
            self.events = sample.info.special_events
        else:
            self.state = None
            self.out_of_sync = None
            self.tick_id = None
            self.timestamp = None
            self.events = None

        self._raw_sample = sample

    def _get_actor_index(self, actor_name):
        if self._actor_indexes is None:
            indexes = {}
            if self._log_params:
                for index, actor in enumerate(self._log_params._raw_params.actors):
                    indexes[actor.name] = index
            else:
                for index, actor in enumerate(self._parameters.actors):
                    indexes[actor.name] = index
            self._actor_indexes = indexes

        return self._actor_indexes.get(actor_name)

    def all_actor_names(self):
        if self._log_params:
            for index in range(self._nb_actors):
                yield self._log_params.get_actor_name(index)
        else:
            for actor in self._parameters.actors:
                yield actor.name

    def get_action(self, actor):
        if isinstance(actor, int):
            actor_index = actor
        elif isinstance(actor, str):
            actor_index = self._get_actor_index(actor)
        else:
            raise CogmentError(f"Wrong type of actor parameter [{type(actor)}]: must be int or str")
        if actor_index is None or actor_index < 0 or actor_index >= self._nb_actors:
            raise CogmentError(f"Invalid actor [{actor}] [{actor_index}]")

        if len(self._raw_sample.actions) > 0:
            data = self._raw_sample.actions[actor_index]
            if actor_index in self._raw_sample.unavailable_actors:
                status = ActorStatus.UNAVAILABLE
                timestamp = 0
                action_space = None
            elif actor_index in self._raw_sample.default_actors:
                status = ActorStatus.DEFAULT
                timestamp = 0
                action_space = None
            else:
                status = ActorStatus.ACTIVE
                timestamp = data.timestamp
                if self._log_params:
                    action_space = self._log_params.get_actor(actor_index)["action_space"]()
                else:
                    action_space = self._parameters.actors[actor_index]._actor_class.action_space()
                action_space.ParseFromString(data.content)

            return RecvAction(actor_index, data.tick_id, status, timestamp, action_space)

        else:
            return None

    def get_observation(self, actor):
        if isinstance(actor, int):
            actor_index = actor
        elif isinstance(actor, str):
            actor_index = self._get_actor_index(actor)
        else:
            raise CogmentError(f"Wrong type of actor parameter [{type(actor)}]: must be int or str")
        if actor_index is None or actor_index < 0 or actor_index >= self._nb_actors:
            raise CogmentError(f"Invalid actor [{actor}] [{actor_index}]")

        if self._log_params:
            obs_space = self._log_params.get_actor(actor_index)["observation_space"]()
        else:
            obs_space = self._parameters.actors[actor_index]._actor_class.observation_space()

        if self._raw_sample.HasField("observations"):
            data = self._raw_sample.observations
            obs_index = data.actors_map[actor_index]
            obs_content = data.observations[obs_index]
            obs_space.ParseFromString(obs_content)
            return RecvObservation(data, obs_space)
        else:
            return None

    def all_rewards(self):
        for rew in self._raw_sample.rewards:
            yield RecvReward(rew)

    def all_messages(self):
        for msg in self._raw_sample.messages:
            yield RecvMessage(msg)


async def _read_sample(context, session, settings):
    try:
        while True:
            request = await context.read()

            if request == grpc.aio.EOF:
                logger.info(f"The orchestrator disconnected from LogExpoterService.")
                break

            elif request.HasField("sample"):
                trial_ended = (request.sample.info.state == common_api.TrialState.ENDED)
                sample = LogSample(session.trial_parameters)
                sample._set(request.sample)
                session._new_sample(sample)
                if trial_ended:
                    logger.debug("Last log sample received for trial")
                    break
            else:
                logger.warning(f"Invalid request received from the orchestrator : {request}")

    except asyncio.CancelledError as exc:
        logger.debug(f"DatalogServicer '_read_sample' coroutine cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_read_sample")
        raise

    # Exit the loop
    session._new_sample(None)


class DatalogServicer(datalog_grpc_api.DatalogSPServicer):
    """Internal datalog servicer class."""

    def __init__(self, impl, cog_settings):
        self._impl = impl
        self._cog_settings = cog_settings
        logger.info("Datalog Service started")

    # Override
    async def RunTrialDatalog(self, request_iterator, context):
        reader_task = None
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]
            user_id = metadata["user-id"]

            request = await context.read()

            if not request.HasField("trial_params"):
                raise CogmentError(f"Initial data log request for [{trial_id}] does not contain parameters.")

            trial_parameters = TrialParameters(None)
            trial_parameters._set(self._cog_settings, request.trial_params)

            session = DatalogSession(self._impl, trial_id, user_id, trial_parameters)
            user_task = session._start_user_task()

            reader_task = asyncio.create_task(_read_sample(context, session, self._cog_settings))

            normal_return = await user_task

            if normal_return:
                logger.debug(f"User datalog implementation returned")
            else:
                logger.debug(f"User datalog implementation was cancelled")

        except asyncio.CancelledError as exc:
            logger.debug(f"Datalog implementation coroutine cancelled: [{exc}]")

        except Exception:
            logger.exception("RunTrialDatalog")
            raise

        finally:
            if reader_task is not None:
                reader_task.cancel()

    # Override
    async def Version(self, request, context):
        try:
            return list_versions()
        except Exception:
            logger.exception("Version")
            raise



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.common_pb2 as common_api
from cogment.utils import logger
from cogment.errors import CogmentError

import asyncio


# Deprecated
class LogParams:
    """Class representing the parameters of a trial for the datalog service."""

    def __init__(self, cog_settings):
        self.max_steps = None
        self.max_inactivity = None
        self.datalog = None
        self.environment = None
        self.nb_actors = None

        self._actor_indexes = None
        self._raw_params = None
        self._cog_settings = cog_settings

    def __str__(self):
        result = f"LogParams: {self._raw_params}"
        return result

    # Type of serialized data being produced and consumed by this class.
    # This is dependent on all the underlying protobuf messages used to
    # serialize/deserialize, and should be incremented if any of them changes in
    # a backward or forward incompatible way. API 1.0 could be considered a type 0 or 1.
    # Current dependencies: TrialParams, DatalogParams, EnvironmentParams, ActorParams,
    #                       SerializedMessage
    def get_serialization_type(self):
        return 2

    def _set(self, raw_params):
        self._set_from_params(raw_params)

    def serialize(self):
        if self._raw_params is None:
            raise CogmentError("Not set, cannot serialize")
        return self._raw_params.SerializeToString()

    def deserialize(self, raw_string):
        params = common_api.TrialParams()
        params.ParseFromString(raw_string)
        self._set_from_params(params)

    def _set_from_params(self, raw_params):
        if type(raw_params) != common_api.TrialParams:
            raise CogmentError(f"Wrong type of parameters provided [{type(raw_params)}]")

        self.max_steps = raw_params.max_steps
        self.max_inactivity = raw_params.max_inactivity

        self.datalog = {
            "endpoint": raw_params.datalog.endpoint,
            "exclude": raw_params.datalog.exclude_fields,
        }

        self.environment = {
            "name": raw_params.environment.name,
            "endpoint": raw_params.environment.endpoint,
            "implementation": raw_params.environment.implementation,
        }

        self.nb_actors = len(raw_params.actors)
        self._actor_indexes = None
        self._raw_params = raw_params

    def get_trial_config(self):
        config = None
        if self._raw_params.HasField("trial_config"):
            config = self._cog_settings.trial.config_type()
            config.ParseFromString(self._raw_params.trial_config.content)

        return config

    def get_environment_config(self):
        config = None
        if(self._raw_params.environment.HasField("config")):
            config = self._cog_settings.environment.config_type()
            config.ParseFromString(self._raw_params.environment.config.content)

        return config

    def get_actor_index(self, actor_name):
        if self._actor_indexes is not None:
            return self._actor_indexes.get(actor_name)
        else:
            indexes = {}
            for index, actor in enumerate(self._raw_params.actors):
                indexes[actor.name] = index
            self._actor_indexes = indexes
            return indexes.get(actor_name)

    def get_actor_name(self, actor_index):
        return self._raw_params.actors[actor_index].name

    def get_actor(self, actor_index):
        actor = self._raw_params.actors[actor_index]

        actor_config = None
        a_c = self._cog_settings.actor_classes.__getattribute__(actor.actor_class)
        if actor.HasField("config"):
            actor_config = a_c.config_type()
            actor_config.ParseFromString(actor.config.content)

        actor_data = {
            "name": actor.name,
            "actor_class": actor.actor_class,
            "endpoint": actor.endpoint,
            "implementation": actor.implementation,
            "config": actor_config,
            "action_space": a_c.action_space,
            "observation_space": a_c.observation_space,
        }

        return actor_data


class DatalogSession:
    """Class representing the session of a datalog for a trial."""

    def __init__(self, impl, trial_id, user_id, trial_parameters):
        self.trial_id = trial_id
        self.user_id = user_id
        self.trial_parameters = trial_parameters

        self._log_params = None
        self._user_task = None
        self._impl = impl
        self._queue = None

    def __str__(self):
        result = f"DatalogSession: trial_id = {self.trial_id}, user_id = {self.user_id}"
        result += f", trial_parameters = {self.trial_parameters}"
        return result

    def start(self):
        self._queue = asyncio.Queue()

    async def all_samples(self):
        if self._queue is not None:
            while True:
                try:
                    sample = await self._queue.get()
                    if sample is None:
                        break
                    keep_looping = yield sample
                    self._queue.task_done()
                    if keep_looping is not None and not bool(keep_looping):
                        break

                except asyncio.CancelledError as exc:
                    logger.debug(f"Datalog coroutine cancelled while waiting for a sample: [{exc}]")
                    break

    async def get_all_samples(self):
        logger.deprecated("'get_all_samples' is deprecated. Use 'all_samples' instead.")

        if self._queue is not None:
            while True:
                try:
                    sample = await self._queue.get()
                    if sample is None:
                        break
                    keep_looping = yield sample
                    self._queue.task_done()
                    if keep_looping is not None and not bool(keep_looping):
                        break

                except GeneratorExit:
                    raise

                except asyncio.CancelledError as exc:
                    logger.debug(f"Datalog coroutine cancelled while waiting for a sample: [{exc}]")
                    break

    def _new_sample(self, sample):
        if self._queue is not None:
            self._queue.put_nowait(sample)
        else:
            logger.warning("Datalog received a sample that it was unable to handle.")

    async def _run(self):
        try:
            await self._impl(self)
            return True

        except asyncio.CancelledError as exc:
            logger.debug(f"Datalog implementation coroutine cancelled: [{exc}]")
            return False

        except Exception:
            logger.exception(f"An exception occured in user datalog implementation:")
            raise

    def _start_user_task(self):
        self._user_task = asyncio.create_task(self._run())
        return self._user_task

    @property
    def trial_params(self):
        if self._log_params is None:
            logger.deprecated(f"DatalogSession's trial_params is deprecated")
            self._log_params = LogParams(self.trial_parameters._cog_settings)
            self._log_params._set(self.trial_parameters._raw_params)

        return self._log_params



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/orchestrator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import environment_pb2 as cogment_dot_api_dot_environment__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x63ogment/api/orchestrator.proto\x12\ncogmentAPI\x1a\x18\x63ogment/api/common.proto\x1a\x1d\x63ogment/api/environment.proto\"\xaa\x01\n\x11TrialStartRequest\x12/\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x1d.cogmentAPI.SerializedMessageH\x00\x12)\n\x06params\x18\x04 \x01(\x0b\x32\x17.cogmentAPI.TrialParamsH\x00\x12\x0f\n\x07user_id\x18\x02 \x01(\t\x12\x1a\n\x12trial_id_requested\x18\x03 \x01(\tB\x0c\n\nstart_data\"#\n\x0fTrialStartReply\x12\x10\n\x08trial_id\x18\x01 \x01(\t\"1\n\x15TerminateTrialRequest\x12\x18\n\x10hard_termination\x18\x01 \x01(\x08\"\x15\n\x13TerminateTrialReply\"J\n\x10TrialInfoRequest\x12\x1e\n\x16get_latest_observation\x18\x01 \x01(\x08\x12\x16\n\x0eget_actor_list\x18\x02 \x01(\x08\"\xe8\x01\n\tTrialInfo\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12%\n\x05state\x18\x02 \x01(\x0e\x32\x16.cogmentAPI.TrialState\x12\x10\n\x08\x65nv_name\x18\x03 \x01(\t\x12\x0f\n\x07tick_id\x18\x04 \x01(\x04\x12\x16\n\x0etrial_duration\x18\x05 \x01(\x06\x12\x36\n\x12latest_observation\x18\x06 \x01(\x0b\x32\x1a.cogmentAPI.ObservationSet\x12/\n\x0f\x61\x63tors_in_trial\x18\x07 \x03(\x0b\x32\x16.cogmentAPI.TrialActor\"6\n\x0eTrialInfoReply\x12$\n\x05trial\x18\x01 \x03(\x0b\x32\x15.cogmentAPI.TrialInfo\":\n\x10TrialListRequest\x12&\n\x06\x66ilter\x18\x01 \x03(\x0e\x32\x16.cogmentAPI.TrialState\"I\n\x0eTrialListEntry\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12%\n\x05state\x18\x02 \x01(\x0e\x32\x16.cogmentAPI.TrialState2\x91\x03\n\x10TrialLifecycleSP\x12J\n\nStartTrial\x12\x1d.cogmentAPI.TrialStartRequest\x1a\x1b.cogmentAPI.TrialStartReply\"\x00\x12V\n\x0eTerminateTrial\x12!.cogmentAPI.TerminateTrialRequest\x1a\x1f.cogmentAPI.TerminateTrialReply\"\x00\x12J\n\x0cGetTrialInfo\x12\x1c.cogmentAPI.TrialInfoRequest\x1a\x1a.cogmentAPI.TrialInfoReply\"\x00\x12K\n\x0bWatchTrials\x12\x1c.cogmentAPI.TrialListRequest\x1a\x1a.cogmentAPI.TrialListEntry\"\x00\x30\x01\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x32\xa4\x01\n\rClientActorSP\x12Q\n\x08RunTrial\x12\x1f.cogmentAPI.ActorRunTrialOutput\x1a\x1e.cogmentAPI.ActorRunTrialInput\"\x00(\x01\x30\x01\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x42\x30Z.github.com/cogment/cogment/grpcapi/cogment/apib\x06proto3')



_TRIALSTARTREQUEST = DESCRIPTOR.message_types_by_name['TrialStartRequest']
_TRIALSTARTREPLY = DESCRIPTOR.message_types_by_name['TrialStartReply']
_TERMINATETRIALREQUEST = DESCRIPTOR.message_types_by_name['TerminateTrialRequest']
_TERMINATETRIALREPLY = DESCRIPTOR.message_types_by_name['TerminateTrialReply']
_TRIALINFOREQUEST = DESCRIPTOR.message_types_by_name['TrialInfoRequest']
_TRIALINFO = DESCRIPTOR.message_types_by_name['TrialInfo']
_TRIALINFOREPLY = DESCRIPTOR.message_types_by_name['TrialInfoReply']
_TRIALLISTREQUEST = DESCRIPTOR.message_types_by_name['TrialListRequest']
_TRIALLISTENTRY = DESCRIPTOR.message_types_by_name['TrialListEntry']
TrialStartRequest = _reflection.GeneratedProtocolMessageType('TrialStartRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialStartRequest)
  })
_sym_db.RegisterMessage(TrialStartRequest)

TrialStartReply = _reflection.GeneratedProtocolMessageType('TrialStartReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialStartReply)
  })
_sym_db.RegisterMessage(TrialStartReply)

TerminateTrialRequest = _reflection.GeneratedProtocolMessageType('TerminateTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TerminateTrialRequest)
  })
_sym_db.RegisterMessage(TerminateTrialRequest)

TerminateTrialReply = _reflection.GeneratedProtocolMessageType('TerminateTrialReply', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TerminateTrialReply)
  })
_sym_db.RegisterMessage(TerminateTrialReply)

TrialInfoRequest = _reflection.GeneratedProtocolMessageType('TrialInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialInfoRequest)
  })
_sym_db.RegisterMessage(TrialInfoRequest)

TrialInfo = _reflection.GeneratedProtocolMessageType('TrialInfo', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFO,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialInfo)
  })
_sym_db.RegisterMessage(TrialInfo)

TrialInfoReply = _reflection.GeneratedProtocolMessageType('TrialInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialInfoReply)
  })
_sym_db.RegisterMessage(TrialInfoReply)

TrialListRequest = _reflection.GeneratedProtocolMessageType('TrialListRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALLISTREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialListRequest)
  })
_sym_db.RegisterMessage(TrialListRequest)

TrialListEntry = _reflection.GeneratedProtocolMessageType('TrialListEntry', (_message.Message,), {
  'DESCRIPTOR' : _TRIALLISTENTRY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.TrialListEntry)
  })
_sym_db.RegisterMessage(TrialListEntry)

_TRIALLIFECYCLESP = DESCRIPTOR.services_by_name['TrialLifecycleSP']
_CLIENTACTORSP = DESCRIPTOR.services_by_name['ClientActorSP']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z.github.com/cogment/cogment/grpcapi/cogment/api'
  _TRIALSTARTREQUEST._serialized_start=104
  _TRIALSTARTREQUEST._serialized_end=274
  _TRIALSTARTREPLY._serialized_start=276
  _TRIALSTARTREPLY._serialized_end=311
  _TERMINATETRIALREQUEST._serialized_start=313
  _TERMINATETRIALREQUEST._serialized_end=362
  _TERMINATETRIALREPLY._serialized_start=364
  _TERMINATETRIALREPLY._serialized_end=385
  _TRIALINFOREQUEST._serialized_start=387
  _TRIALINFOREQUEST._serialized_end=461
  _TRIALINFO._serialized_start=464
  _TRIALINFO._serialized_end=696
  _TRIALINFOREPLY._serialized_start=698
  _TRIALINFOREPLY._serialized_end=752
  _TRIALLISTREQUEST._serialized_start=754
  _TRIALLISTREQUEST._serialized_end=812
  _TRIALLISTENTRY._serialized_start=814
  _TRIALLISTENTRY._serialized_end=887
  _TRIALLIFECYCLESP._serialized_start=890
  _TRIALLIFECYCLESP._serialized_end=1291
  _CLIENTACTORSP._serialized_start=1294
  _CLIENTACTORSP._serialized_end=1458
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.utils import logger

import os
from typing import List, Tuple

# Schemes
GRPC_SCHEME = "grpc"
COGMENT_SCHEME = "cogment"

# Hosts
DISCOVERY_HOST = "discover"
CLIENT_ACTOR_HOST = "client"

# Paths
SERVICE_PATH = "service"
ACTOR_PATH = "actor"
ENVIRONMENT_PATH = "environment"
DATALOG_PATH = "datalog"
PRE_HOOK_PATH = "prehook"
LIFECYCLE_PATH = "lifecycle"
ACTOR_SERVICE_PATH = "actservice"
DATASTORE_PATH = "datastore"
MODEL_REGISTRY_PATH = "modelregistry"

# Reserved query property names
ACTOR_CLASS_PROPERTY_NAME = "__actor_class"
IMPLEMENTATION_PROPERTY_NAME = "__implementation"
SERVICE_ID_PROPERTY_NAME = "__id"
AUTHENTICATION_TOKEN_PROPERTY_NAME = "__authentication-token"

# Special URLs
BASIC_CONTEXT_DISCOVERY_URL = "cogment://discover"
CLIENT_ACTOR_URL = "cogment://client"


class Endpoint:
    """Class representing a remote Cogment endpoint where to connect."""

    def __init__(self, url: str = BASIC_CONTEXT_DISCOVERY_URL):
        self.url: str = url
        self.private_key: str = None
        self.root_certificates: str = None
        self.certificate_chain: str = None

    def __str__(self):
        result = f"Endpoint: url = {self.url}, private_key = {self.private_key}"
        result += f", root_certificates = {self.root_certificates}, certificate_chain = {self.certificate_chain}"
        return result

    def using_ssl(self):
        return (self.private_key is not None)

    def set_from_files(self, private_key_file=None, root_certificates_file=None, certificate_chain_file=None):
        try:
            if private_key_file:
                with open(private_key_file) as fl:
                    self.private_key = fl.read()
                if not self.private_key:
                    self.private_key = None

            if root_certificates_file:
                with open(root_certificates_file) as fl:
                    self.root_certificates = fl.read()
                if not self.root_certificates:
                    self.root_certificates = None

            if certificate_chain_file:
                with open(certificate_chain_file) as fl:
                    self.certificate_chain = fl.read()
                if not self.certificate_chain:
                    self.certificate_chain = None

        except Exception:
            logger.error(f"Failed loading file from CWD: {os.getcwd()}")
            raise


class ServedEndpoint:
    """Class representing a local Cogment endpoint where others can connect."""

    def __init__(self, port: int):
        if type(port) == str:
            logger.deprecated(f"'ServedEndpoint' port should be an integer, not a string")  # type: ignore
            self.port = int(port)
        else:
            self.port = port
        self.private_key_certificate_chain_pairs: List[Tuple[str, str]] = None
        self.root_certificates: str = None

    def using_ssl(self):
        return (self.private_key_certificate_chain_pairs is not None)

    # def set_from_files(private_key_certificate_chain_pairs_file=None, root_certificates_file=None):
    # TODO: This function would need to parse the PEM encoded `private_key_certificate_chain_pairs_file`
    #       to create the list of tuples required (see simpler version in `Endpoint` class above).

    def __str__(self):
        result = f"ServedEndpoint: port = {self.port}"
        result += f", private_key_certificate_chain_pairs = {self.private_key_certificate_chain_pairs}"
        result += f", root_certificates = {self.root_certificates}"
        return result



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc.aio  # type: ignore
from prometheus_client import Summary, Counter

import cogment.api.environment_pb2_grpc as env_grpc_api
import cogment.api.environment_pb2 as env_api
import cogment.api.common_pb2 as common_api

from cogment.utils import list_versions, logger, INIT_TIMEOUT
from cogment.session import RecvEvent, RecvMessage, ActorStatus, RecvAction, EventType
from cogment.session import _InitAck, _EndingAck, _Ending
from cogment.errors import CogmentError
from cogment.environment import EnvironmentSession
from cogment.trial import Trial

import asyncio
import time


class _PrometheusData:
    """Internal class holding the details of Prometheus report values for an environment."""

    def __init__(self, prometheus_registry):
        self.update_count_per_trial = Summary(
            "environment_update_count_per_trial",
            "Number of update by trial",
            ["impl_name"],
            registry=prometheus_registry
        )
        self.trial_duration = Summary(
            "environment_trial_duration_in_second", "Trial duration", ["trial_actor"],
            registry=prometheus_registry
        )
        self.trials_started = Counter(
            "environment_trials_started", "Number of trial started", ["impl_name"],
            registry=prometheus_registry
        )
        self.trials_ended = Counter(
            "environment_trials_ended", "Number of trial ended", ["impl_name"],
            registry=prometheus_registry
        )
        self.messages_received = Counter(
            "environment_received_messages",
            "Number of messages received",
            ["impl_name"],
            registry=prometheus_registry
        )
        self.messages_sent = Counter(
            "environment_sent_messages",
            "Number of messages sent",
            ["impl_name"],
            registry=prometheus_registry
        )

        pass


def _trial_key(trial_id, actor_name):
    return f"{trial_id}_{actor_name}"


def _process_normal_data(data, session):
    if session._trial.ending:
        recv_event = RecvEvent(EventType.ENDING)
    else:
        recv_event = RecvEvent(EventType.ACTIVE)

    if data.HasField("action_set"):
        logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}] received action set")

        act_set = data.action_set

        len_actions = len(act_set.actions)
        len_actors = len(session._trial.actors)

        if len_actions != len_actors:
            raise CogmentError(f"Received {len_actions} actions but have {len_actors} actors")

        tick_id = act_set.tick_id
        session._trial.tick_id = tick_id

        for index, actor in enumerate(session._trial.actors):
            if index in act_set.unavailable_actors:
                status = ActorStatus.UNAVAILABLE
                timestamp = 0
                action = None
            else:
                status = ActorStatus.ACTIVE
                timestamp = act_set.timestamp
                action = actor.actor_class.action_space()
                action.ParseFromString(act_set.actions[index])

            recv_event.actions.append(RecvAction(index, act_set.tick_id, status, timestamp, action))

        session._post_incoming_event((tick_id, recv_event))

    elif data.HasField("message"):
        logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}] received message")

        recv_event.messages = [RecvMessage(data.message)]
        session._post_incoming_event((-1, recv_event))

        session._prometheus_data.messages_received.labels(session.impl_name).inc()

    elif data.HasField("details"):
        logger.warning(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                       f"received unexpected detail data [{data.details}]")

    else:
        logger.error(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                     f"received unexpected data [{data.WhichOneof('data')}]")


async def _process_incoming(context, session):
    try:
        while True:
            data = await context.read()
            if data == grpc.aio.EOF:
                logger.info(f"Trial [{session._trial.id}] - Environment [{session.name}]: "
                            f"The orchestrator disconnected the environment")
                break
            logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}]: "
                         f"Received data [{data.state}] [{data.WhichOneof('data')}]")

            if data.state == common_api.CommunicationState.NORMAL:
                _process_normal_data(data, session)

            elif data.state == common_api.CommunicationState.HEARTBEAT:
                logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                             f"received 'HEARTBEAT' and responding in kind")
                reply = env_api.EnvRunTrialOutput()
                reply.state = data.state
                await context.write(reply)

            elif data.state == common_api.CommunicationState.LAST:
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                             f"received 'LAST' state")
                session._trial.ending = True

            elif data.state == common_api.CommunicationState.LAST_ACK:
                logger.error(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                             f"received an unexpected 'LAST_ACK'")
                # TODO: Should we `return` or raise instead of continuing?

            elif data.state == common_api.CommunicationState.END:
                if session._trial.ending:
                    if data.HasField("details"):
                        logger.info(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                                    f"ended [{data.details}]")
                    else:
                        logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] ended")
                    session._post_incoming_event((-1, RecvEvent(EventType.FINAL)))
                else:
                    if data.HasField("details"):
                        details = data.details
                    else:
                        details = ""
                    logger.warning(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                                   f"ended forcefully [{details}]")

                session._trial.ended = True
                session._exit_queues()
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                             f"received an invalid state [{data.state}]")

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                     f"process incoming cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_process_incoming")
        raise


async def _process_outgoing(context, session):
    try:
        async for data in session._retrieve_outgoing_data():
            package = env_api.EnvRunTrialOutput()
            package.state = common_api.CommunicationState.NORMAL

            # Using strict comparison: there is no reason to receive derived classes here
            if type(data) == env_api.ObservationSet:
                logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}]: "
                             f"Sending observation set")
                package.observation_set.CopyFrom(data)

            elif type(data) == common_api.Reward:
                logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}]: Sending reward")
                package.reward.CopyFrom(data)

            elif type(data) == common_api.Message:
                logger.trace(f"Trial [{session._trial.id}] - Environment [{session.name}]: Sending message")
                session._prometheus_data.messages_sent.labels(session.impl_name).inc()
                package.message.CopyFrom(data)

            elif type(data) == _InitAck:
                package.init_output.SetInParent()
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}]: Sending Init Data")

            elif type(data) == _Ending:
                package.state = common_api.CommunicationState.LAST
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}]: Sending 'LAST'")

            elif type(data) == _EndingAck:
                package.state = common_api.CommunicationState.LAST_ACK
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}]: Sending 'LAST_ACK'")
                await context.write(package)
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Environment [{session.name}]: "
                             f"Unknown data type to send [{type(data)}]")
                continue

            await context.write(package)

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                     f"process outgoing cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_process_outgoing")
        raise


class EnvironmentServicer(env_grpc_api.EnvironmentSPServicer):
    """Internal environment servicer class."""

    def __init__(self, env_impls, cog_settings, prometheus_registry=None):
        self._impls = env_impls
        self._sessions = set()
        self._cog_settings = cog_settings
        self._prometheus_data = _PrometheusData(prometheus_registry)

        logger.info("Environment Service started")

    async def _get_init_data(self, context, trial_id):
        logger.debug(f"Trial [{trial_id}] - Processing init for environment")

        last_received = False
        while True:
            request = await context.read()
            logger.debug(f"Trial [{trial_id}] - Read initial environment request: {request}")

            if request == grpc.aio.EOF:
                logger.info(f"Trial [{trial_id}] - Orchestrator disconnected environment before start")
                return None

            if request.state == common_api.CommunicationState.NORMAL:
                if last_received:
                    error_str = (f"Trial [{trial_id}] - Environment received init data after 'LAST'")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

                if request.HasField("init_input"):
                    return request.init_input
                elif reply.HasField("details"):
                    logger.warning(f"Trial [{trial_id}] - Received unexpected detail data "
                                   f"[{reply.details}] before start")
                else:
                    data_set = request.WhichOneof("data")
                    error_str = (f"Trial [{trial_id}] - Environment received unexpected init data [{data_set}]")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

            elif request.state == common_api.CommunicationState.HEARTBEAT:
                reply = env_api.EnvRunTrialOutput()
                reply.state = request.state
                await context.write(reply)

            elif request.state == common_api.CommunicationState.LAST:
                if last_received:
                    error_str = (f"Trial [{trial_id}] - Before start, environment "
                                 f"received unexpected 'LAST' when waiting for 'END'")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

                logger.debug(f"Trial [{trial_id}] - Ending before init data")
                reply = env_api.EnvRunTrialOutput()
                reply.state = common_api.CommunicationState.LAST_ACK
                await context.write(reply)
                last_received = True

            elif request.state == common_api.CommunicationState.LAST_ACK:
                error_str = (f"Trial [{trial_id}] - Before start, environment received an unexpected 'LAST_ACK'")
                logger.error(error_str)
                await context.abort(grpc.StatusCode.UNKNOWN, error_str)

            elif request.state == common_api.CommunicationState.END:
                if request.HasField("details"):
                    logger.info(f"Trial [{trial_id}] - Ended before environment start [{request.details}]")
                else:
                    logger.debug(f"Trial [{trial_id}] - Ended before environment start")
                return None

            else:
                error_str = (f"Trial [{trial_id}] - Before start, environment "
                             f"received an invalid state [{request.state}] [{request}]")
                logger.error(error_str)
                await context.abort(grpc.StatusCode.UNKNOWN, error_str)

    def _start_session(self, trial_id, init_input):
        name = init_input.name
        if not name:
            raise CogmentError(f"Trial [{trial_id}] - Empty environment name")

        if init_input.impl_name:
            impl_name = init_input.impl_name
            impl = self._impls.get(impl_name)
            if impl is None:
                raise CogmentError(f"Trial [{trial_id}] - "
                                   f"Unknown impl [{impl_name}] for environment [{name}]")
        else:
            impl_name, impl = next(iter(self._impls.items()))
            logger.info(f"Trial [{trial_id}] - "
                        f"impl [{impl_name}] arbitrarily chosen for environment [{name}]")

        key = _trial_key(trial_id, name)
        if key in self._sessions:
            raise CogmentError(f"Trial [{trial_id}] - Environment [{name}] already exists")

        config = None
        if init_input.HasField("config"):
            env_config_type = self._cog_settings.environment.config_type
            if env_config_type is None:
                raise CogmentError(f"Trial [{trial_id}] - Environment [{name}] "
                                   f"received config data of unknown type (was it defined in cogment.yaml?)")
            config = env_config_type()
            config.ParseFromString(init_input.config.content)

        self._prometheus_data.trials_started.labels(impl_name).inc()

        trial = Trial(trial_id, init_input.actors_in_trial, self._cog_settings)
        trial.tick_id = init_input.tick_id
        new_session = EnvironmentSession(impl.impl, trial, name, impl_name, config)
        new_session._prometheus_data = self._prometheus_data
        self._sessions.add(key)

        logger.debug(f"Trial [{trial_id}] - impl [{impl_name}] for environment [{name}] started")

        return new_session

    async def _run_session(self, context, session):
        send_task = None
        process_task = None

        try:
            send_task = asyncio.create_task(_process_outgoing(context, session))
            process_task = asyncio.create_task(_process_incoming(context, session))
            user_task = session._start_user_task()

            with self._prometheus_data.trial_duration.labels(session.impl_name).time():
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] session started")
                normal_return = await user_task

            self._prometheus_data.update_count_per_trial.labels(session.impl_name).observe(session._trial.tick_id)

            if normal_return:
                if not session._last_event_delivered:
                    logger.warning(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                                   f"user implementation returned before required")
                else:
                    logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                                 f"user implementation returned")
            else:
                logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                             f"user implementation was cancelled")

            self._prometheus_data.trials_ended.labels(session.impl_name).inc()

        except asyncio.CancelledError as exc:
            logger.debug(f"Trial [{session._trial.id}] - Environment [{session.name}] "
                         f"user implementation was cancelled with exception [{exc}]")

        except Exception:
            logger.exception("run_session")
            raise

        finally:
            if send_task is not None:
                send_task.cancel()
            if process_task is not None:
                process_task.cancel()

    # Override
    async def RunTrial(self, request_iterator, context):
        if len(self._impls) == 0:
            logger.warning("No implementation registered on trial run request")
            raise CogmentError("No implementation registered")

        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        try:
            init_data = await asyncio.wait_for(self._get_init_data(context, trial_id), INIT_TIMEOUT)
            if init_data is None:
                return
            key = _trial_key(trial_id, init_data.name)

            session = self._start_session(trial_id, init_data)

        except asyncio.TimeoutError:
            logger.error("Failed to receive init data from Orchestrator")

        except grpc._cython.cygrpc.AbortError:  # Exception from context.abort()
            raise

        except asyncio.CancelledError as exc:
            logger.debug(f"Trial [{trial_id}] - Environment start cancelled: [{exc}]")
            raise

        except Exception:
            logger.exception("RunTrial")
            raise

        else:
            await self._run_session(context, session)  # Blocking
            self._sessions.remove(key)

    # Override
    async def Version(self, request, context):
        try:
            return list_versions()

        except Exception:
            logger.exception("Version")
            raise



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc  # type: ignore
import grpc.aio  # type: ignore
from prometheus_client import Summary, Counter, Gauge

import cogment.api.agent_pb2_grpc as agent_grpc_api
import cogment.api.common_pb2 as common_api

from cogment.utils import list_versions, logger, INIT_TIMEOUT
from cogment.trial import Trial
from cogment.actor import ActorSession
from cogment.session import RecvEvent, RecvObservation, RecvMessage, RecvReward, EventType, _InitAck, _EndingAck
from cogment.errors import CogmentError

import asyncio


class _PrometheusData:
    """Internal class holding the details of Prometheus report values for a service actor."""

    def __init__(self, prometheus_registry):
        self.decide_request_time = Summary(
            "actor_decide_processing_seconds",
            "Time spent by an actor on the decide function",
            ["name", "impl_name"],
            registry=prometheus_registry)
        self.actors_started = Counter(
            "actor_started", "Number of actors created", ["impl_name"],
            registry=prometheus_registry)
        self.actors_ended = Counter(
            "actor_ended", "Number of actors ended", ["impl_name"],
            registry=prometheus_registry)
        self.messages_received = Counter(
            "actor_received_messages", "Number of messages received", ["name", "impl_name"],
            registry=prometheus_registry)
        self.rewards_received = Gauge(
            "actor_reward_summation", "Cumulative rewards received", ["name", "impl_name"],
            registry=prometheus_registry)
        self.rewards_counter = Counter(
            "actor_rewards_count", "Number of rewards received", ["name", "impl_name"],
            registry=prometheus_registry)


def _trial_key(trial_id, actor_name):
    return f"{trial_id}_{actor_name}"


def get_actor_impl(trial_id, actor_impls, init_data):
    if not init_data.actor_name:
        raise CogmentError(f"Trial [{trial_id}] - Empty actor name")
    if not init_data.actor_class:
        raise CogmentError(f"Trial [{trial_id}] - Empty actor class for actor [{init_data.actor_name}]")

    actor_impl = actor_impls.get(init_data.impl_name)
    if actor_impl is not None:
        # Check compatibility
        compatible = (len(actor_impl.actor_classes) == 0)
        for class_name in actor_impl.actor_classes:
            if class_name == init_data.actor_class:
                compatible = True
                break
        if not compatible:
            raise CogmentError(f"Trial [{trial_id}] - actor [{init_data.actor_name}] class [{init_data.actor_class}] "
                               f"is not compatible with actor implementation [{init_data.impl_name}]")
    else:
        # Find a compatible actor class in the registered actors

        # Search for exact match first
        for init_data.impl_name, impl in actor_impls.items():
            for class_name in impl.actor_classes:
                if class_name == init_data.actor_class:
                    actor_impl = impl
                    break
            if actor_impl is not None:
                break

        if actor_impl is None:
            raise CogmentError(f"Trial [{trial_id}] - actor [{init_data.actor_name}] class [{init_data.actor_class}] "
                               f"is not compatible with with any registered actor")

        logger.info(f"Trial [{trial_id}] - "
                    f"impl [{init_data.impl_name}] arbitrarily chosen for actor [{init_data.actor_name}]")

    return actor_impl


def _process_normal_data(data, session):
    if session._trial.ending:
        recv_event = RecvEvent(EventType.ENDING)
    else:
        recv_event = RecvEvent(EventType.ACTIVE)

    if data.HasField("observation"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received an observation")

        if session._trial.ending and session._auto_ack:
            session._post_outgoing_data(_EndingAck())

        tick_id = data.observation.tick_id
        session._trial.tick_id = tick_id

        obs_space = session._actor_class.observation_space()
        obs_space.ParseFromString(data.observation.content)

        recv_event.observation = RecvObservation(data.observation, obs_space)
        session._post_incoming_event((tick_id, recv_event))

    elif data.HasField("reward"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received reward")

        recv_event.rewards = [RecvReward(data.reward)]
        session._post_incoming_event((-1, recv_event))

        value = recv_event.rewards[0].value
        session._prometheus_data.rewards_counter.labels(session.name, session.impl_name).inc()
        if value < 0.0:
            session._prometheus_data.rewards_received.labels(session.name, session.impl_name).dec(abs(value))
        else:
            session._prometheus_data.rewards_received.labels(session.name, session.impl_name).inc(value)

    elif data.HasField("message"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received message")

        recv_event.messages = [RecvMessage(data.message)]
        session._post_incoming_event((-1, recv_event))

        session._prometheus_data.messages_received.labels(session.name, session.impl_name).inc()

    elif data.HasField("details"):
        logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                       f"received unexpected detail data [{data.details}]")

    else:
        logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                     f"received unexpected data [{data.WhichOneof('data')}]")


async def _process_incoming(context, session):
    try:
        while True:
            data = await context.read()
            if data == grpc.aio.EOF:
                logger.info(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                            f"The orchestrator disconnected the actor")
                break
            logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                         f"Received data [{data.state}] [{data.WhichOneof('data')}]")

            if data.state == common_api.CommunicationState.NORMAL:
                _process_normal_data(data, session)

            elif data.state == common_api.CommunicationState.HEARTBEAT:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received 'HEARTBEAT' and responding in kind")
                reply = common_api.ActorRunTrialOutput()
                reply.state = data.state
                await context.write(reply)

            elif data.state == common_api.CommunicationState.LAST:
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] received 'LAST' state")
                session._trial.ending = True

            elif data.state == common_api.CommunicationState.LAST_ACK:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received an unexpected 'LAST_ACK'")
                # TODO: Should we `return` or raise instead of continuing?

            elif data.state == common_api.CommunicationState.END:
                if session._trial.ending:
                    if data.HasField("details"):
                        logger.info(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                                    f"Trial ended with explanation [{data.details}]")
                    else:
                        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}]: Trial ended")
                    session._post_incoming_event((-1, RecvEvent(EventType.FINAL)))
                else:
                    if data.HasField("details"):
                        details = data.details
                    else:
                        details = ""
                    logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                                   f"Trial ended forcefully [{details}]")

                session._trial.ended = True
                session._exit_queues()
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received an invalid state [{data.state}]")

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] coroutine cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_process_incoming")
        raise


async def _process_outgoing(context, session):
    try:
        async for data in session._retrieve_outgoing_data():
            package = common_api.ActorRunTrialOutput()
            package.state = common_api.CommunicationState.NORMAL

            # Using strict comparison: there is no reason to receive derived classes here
            if type(data) == common_api.Action:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending action")
                package.action.CopyFrom(data)

            elif type(data) == common_api.Reward:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending reward")
                package.reward.CopyFrom(data)

            elif type(data) == common_api.Message:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending message")
                package.message.CopyFrom(data)

            elif type(data) == _InitAck:
                package.init_output.SetInParent()
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending Init Data")

            elif type(data) == _EndingAck:
                package.state = common_api.CommunicationState.LAST_ACK
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending 'LAST_ACK'")
                await context.write(package)
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                             f"Unknown data type to send [{type(data)}]")
                continue

            await context.write(package)

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] process outgoing cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_process_outgoing")
        raise


class AgentServicer(agent_grpc_api.ServiceActorSPServicer):
    """Internal service actor servicer class."""

    def __init__(self, agent_impls, cog_settings, prometheus_registry=None):
        self._impls = agent_impls
        self._sessions = set()
        self._cog_settings = cog_settings
        self._prometheus_data = _PrometheusData(prometheus_registry)

        logger.info("Agent Service started")

    async def _get_init_data(self, context, trial_id):
        logger.debug(f"Trial [{trial_id}] - Processing init for service actor")

        last_received = False
        while True:
            request = await context.read()
            logger.debug(f"Trial [{trial_id}] - Read initial request: {request}")

            if request == grpc.aio.EOF:
                logger.info(f"Trial [{trial_id}] - Orchestrator disconnected service actor before start")
                return None

            if request.state == common_api.CommunicationState.NORMAL:
                if last_received:
                    error_str = (f"Trial [{trial_id}] - Received init data after 'LAST'")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

                if request.HasField("init_input"):
                    return request.init_input
                elif reply.HasField("details"):
                    logger.warning(f"Trial [{trial_id}] - Received unexpected detail data "
                                   f"[{reply.details}] before start")
                else:
                    data_set = request.WhichOneof("data")
                    error_str = (f"Trial [{trial_id}] - Received unexpected init data [{data_set}]")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

            elif request.state == common_api.CommunicationState.HEARTBEAT:
                reply = common_api.ActorRunTrialOutput()
                reply.state = request.state
                await context.write(reply)

            elif request.state == common_api.CommunicationState.LAST:
                if last_received:
                    error_str = (f"Trial [{trial_id}] - Before start, received unexpected 'LAST' "
                                    f"when waiting for 'END'")
                    logger.error(error_str)
                    await context.abort(grpc.StatusCode.UNKNOWN, error_str)

                logger.debug(f"Trial [{trial_id}] - Ending before init data")
                reply = common_api.ActorRunTrialOutput()
                reply.state = common_api.CommunicationState.LAST_ACK
                await context.write(reply)
                last_received = True

            elif request.state == common_api.CommunicationState.LAST_ACK:
                error_str = (f"Trial [{trial_id}] - Before start, received an unexpected 'LAST_ACK'")
                logger.error(error_str)
                await context.abort(grpc.StatusCode.UNKNOWN, error_str)

            elif request.state == common_api.CommunicationState.END:
                if request.HasField("details"):
                    logger.info(f"Trial [{trial_id}] - Ended before start [{request.details}]")
                else:
                    logger.debug(f"Trial [{trial_id}] - Ended before start")
                return None

            else:
                error_str = (f"Trial [{trial_id}] - Before start, received an invalid state "
                                f"[{request.state}] [{request}]")
                logger.error(error_str)
                await context.abort(grpc.StatusCode.UNKNOWN, error_str)

    def _start_session(self, trial_id, init_input):
        actor_impl = get_actor_impl(trial_id, self._impls, init_input)

        actor_name = init_input.actor_name
        actor_class = self._cog_settings.actor_classes.get(init_input.actor_class)
        if actor_class is None:
            raise CogmentError(f"Trial [{trial_id}] - "
                               f"Unknown class [{init_input.actor_class}] for service actor [{actor_name}]")

        key = _trial_key(trial_id, actor_name)
        if key in self._sessions:
            raise CogmentError(f"Trial [{trial_id}] - Service actor [{actor_name}] already exists")

        config = None
        if init_input.HasField("config"):
            if actor_class.config_type is None:
                raise CogmentError(f"Trial [{trial_id}] - Service actor [{actor_name}] "
                                   f"received config data of unknown type (was it defined in cogment.yaml?)")
            config = actor_class.config_type()
            config.ParseFromString(init_input.config.content)

        self._prometheus_data.actors_started.labels(init_input.impl_name).inc()

        trial = Trial(trial_id, [], self._cog_settings)
        new_session = ActorSession(actor_impl.impl, actor_class, trial, actor_name, init_input.impl_name,
                                   init_input.env_name, config)
        new_session._prometheus_data = self._prometheus_data
        self._sessions.add(key)

        logger.debug(f"Trial [{trial_id}] - impl [{init_input.impl_name}] for service actor [{actor_name}] started")

        return new_session

    async def _run_session(self, context, session):
        send_task = None
        process_task = None

        try:
            send_task = asyncio.create_task(_process_outgoing(context, session))
            process_task = asyncio.create_task(_process_incoming(context, session))
            user_task = session._start_user_task()

            with self._prometheus_data.decide_request_time.labels(session.name, session.impl_name).time():
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] session started")
                normal_return = await user_task

            if normal_return:
                if not session._last_event_delivered:
                    logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                                   f"user implementation returned before required")
                else:
                    logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                                 f"user implementation returned")
            else:
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"user implementation was cancelled")

            self._prometheus_data.actors_ended.labels(session.impl_name).inc()

        except asyncio.CancelledError as exc:
            logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                         f"user implementation was cancelled")

        except Exception:
            logger.exception("run_session")
            raise

        finally:
            if send_task is not None:
                send_task.cancel()
            if process_task is not None:
                process_task.cancel()

    # Override
    async def RunTrial(self, request_iterator, context):
        if len(self._impls) == 0:
            logger.warning("No implementation registered on trial run request")
            raise CogmentError("No implementation registered")

        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        try:
            init_data = await asyncio.wait_for(self._get_init_data(context, trial_id), INIT_TIMEOUT)
            if init_data is None:
                return
            key = _trial_key(trial_id, init_data.actor_name)

            session = self._start_session(trial_id, init_data)

            await self._run_session(context, session)  # Blocking

            self._sessions.remove(key)

        except asyncio.TimeoutError:
            logger.error("Failed to receive init data from Orchestrator")

        except grpc._cython.cygrpc.AbortError:  # Exception from context.abort()
            pass

        except Exception:
            logger.exception("RunTrial")
            raise

    # Override
    async def Version(self, request, context):
        try:
            return list_versions()

        except Exception:
            logger.exception("Version")
            raise



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.aio  # type: ignore

import cogment.api.directory_pb2 as directory_api

from cogment.errors import CogmentError
from cogment.utils import logger
import cogment.endpoints as ep

import copy
import urllib.parse as urlpar
from enum import Enum
from typing import Dict, List, Tuple

AUTHENTICATION_TOKEN_METADATA_NAME = "authentication-token"


class ServiceType(Enum):
    """Enum class for the different services available in directory."""

    LIFE_CYCLE = directory_api.ServiceType.TRIAL_LIFE_CYCLE_SERVICE
    CLIENT_ACTOR = directory_api.ServiceType.CLIENT_ACTOR_CONNECTION_SERVICE
    ACTOR = directory_api.ServiceType.ACTOR_SERVICE
    ENVIRONMENT = directory_api.ServiceType.ENVIRONMENT_SERVICE
    HOOK = directory_api.ServiceType.PRE_HOOK_SERVICE
    DATALOG = directory_api.ServiceType.DATALOG_SERVICE
    DATASTORE = directory_api.ServiceType.DATASTORE_SERVICE
    MODEL_REG = directory_api.ServiceType.MODEL_REGISTRY_SERVICE
    OTHER = directory_api.ServiceType.OTHER_SERVICE


# TODO: Add a cache to service?
# TODO: Make ready for end users?
class Directory:

    def __init__(self, stub, auth_token: str):
        self._stub = stub
        self._auth_token = auth_token

        if self._auth_token:
            self._metadata = [(AUTHENTICATION_TOKEN_METADATA_NAME, self._auth_token)]
        else:
            self._metadata = []

    def __str__(self):
        result = f"Directory: authentication token = {self._auth_token}"
        return result

    async def register_host(self, type: ServiceType, host: str, port: int, ssl=False,
                            properties: Dict[str, str] = None):
        request = directory_api.RegisterRequest()
        if ssl:
            request.endpoint.protocol = directory_api.ServiceEndpoint.Protocol.GRPC_SSL
        else:
            request.endpoint.protocol = directory_api.ServiceEndpoint.Protocol.GRPC
        request.endpoint.host = host
        request.endpoint.port = port
        request.details.type = type.value
        if properties is not None:
            request.details.properties.update(properties)

        def request_itor():
            yield request

        reply_itor = self._stub.Register(request_itor(), metadata=self._metadata)
        if not reply_itor:
            raise CogmentError("Failed to connect to the directory")

        async for reply in reply_itor:
            if reply == grpc.aio.EOF:
                raise CogmentError(f"No response to directory register request [{request}]")

            if reply.status != directory_api.RegisterReply.Status.OK:
                raise CogmentError(f"Failed to register to the directory [{reply.status}] [{reply.error_msg}]")

            logger.debug(f"New directory registration [{type}] [{host}] [{port}] [{ssl}] [{properties}]"
                         f" service id [{reply.service_id}]")
            return (reply.service_id, reply.secret)

        raise CogmentError(f"Empty response to directory register request [{request}]")

    async def deregister_service(self, service_id: int, secret: str):
        request = directory_api.DeregisterRequest()
        request.service_id = service_id
        request.secret = secret

        def request_itor():
            yield request

        reply_itor = self._stub.Deregister(request_itor(), metadata=self._metadata)
        if not reply_itor:
            raise CogmentError("Failed to connect to the directory")

        async for reply in reply_itor:
            if reply == grpc.aio.EOF:
                CogmentError(f"No response to deregister request")

            if reply.status != directory_api.DeregisterReply.Status.OK:
                CogmentError(f"Deregistration failed [{reply.status}] [{reply.error_msg}]")

            logger.debug(f"Deregistered from directory: service id [{service_id}]")

    async def inquire_by_id(self, service_id: int):
        request = directory_api.InquireRequest()
        request.service_id = service_id
        dir_urls = await self._inquire(request)

        return dir_urls

    async def inquire_by_type(self, type: ServiceType, properties: Dict[str, str]):
        request = directory_api.InquireRequest()
        request.details.type = type.value
        for key, val in properties.items():
            request.details.properties[key] = val
        dir_urls = await self._inquire(request)

        return dir_urls

    async def _inquire(self, request) -> List[Tuple[str, bool]]:
        result: List[Tuple[str, bool]] = []

        reply_itor = self._stub.Inquire(request, metadata=self._metadata)
        if not reply_itor:
            raise CogmentError("Failed to connect to the directory")

        async for reply in reply_itor:
            if reply == grpc.aio.EOF:
                logger.debug(f"No response to directory inquire request [{request}]")
                return result

            if reply.data.endpoint.protocol == directory_api.ServiceEndpoint.Protocol.GRPC:
                address = f"{ep.GRPC_SCHEME}://{reply.data.endpoint.host}:{reply.data.endpoint.port}"
                ssl = False
            elif reply.data.endpoint.protocol == directory_api.ServiceEndpoint.Protocol.GRPC_SSL:
                address = f"{ep.GRPC_SCHEME}://{reply.data.endpoint.host}:{reply.data.endpoint.port}"
                ssl = True
            elif reply.data.endpoint.protocol == directory_api.ServiceEndpoint.Protocol.COGMENT:
                host = reply.data.endpoint.host
                if host != ep.CLIENT_ACTOR_HOST:
                    raise CogmentError(f"Invalid cogment endpoint from directory [{host}]")
                address = ep.CLIENT_ACTOR_URL
                ssl = False
            else:
                raise CogmentError(f"Invalid reply from directory inquiry [{reply}]")

            result.append((address, ssl))

        return result

    def _new_endpoint(self, endpoint: ep.Endpoint, dir_urls):
        if len(dir_urls) == 0:
            # TODO: Decide what to do in case inquiry did not find a service in directory
            return endpoint

        url = None
        ssl_endpoint = endpoint.using_ssl()
        for dir_url, dir_ssl in dir_urls:
            if (ssl_endpoint and dir_ssl) or (not ssl_endpoint and not dir_ssl):
                url = dir_url

        if url is not None:
            new_endpoint = copy.copy(endpoint)
            new_endpoint.url = url
        else:
            if not ssl_endpoint:
                raise CogmentError(f"All services found require an SSL connection [{endpoint.url}]")
            new_endpoint = ep.Endpoint(dir_urls[0][0])

        return new_endpoint

    # The URL data takes precedence over the provided parameters
    async def get_inquired_endpoint(self, endpoint: ep.Endpoint, type: ServiceType = None,
                                    properties: Dict[str, str] = None):
        """ Inquire directory for an enpoint with optional context (type and properties)"""

        try:
            parsed_url = urlpar.urlparse(endpoint.url)
        except Exception as exc:
            raise CogmentError(f"Endpoint [{endpoint.url}]: {exc}")

        if parsed_url.scheme == ep.GRPC_SCHEME:
            logger.debug(f"No need to inquire directory, endpoint is already a grpc endpoint")
            return endpoint

        if parsed_url.scheme != ep.COGMENT_SCHEME:
            raise CogmentError(
                f"Invalid endpoint scheme (must be '{ep.GRPC_SCHEME}' or '{ep.COGMENT_SCHEME}') [{endpoint.url}]")
        if parsed_url.hostname != ep.DISCOVERY_HOST:
            raise CogmentError(f"Unknown host for cogment endpoint [{endpoint.url}]")

        path = parsed_url.path.strip("/")
        if len(path) == 0:
            if type is None:
                raise CogmentError(f"No type provided for inquiry from directory")
            dir_type = type
        else:
            if type is not None:
                logger.warning(f"Parameter 'type' [{type}] is ignored due to a path being provided [{endpoint.url}]")

            if path == ep.SERVICE_PATH:
                # properties are ignored
                if parsed_url.query:
                    try:
                        query = urlpar.parse_qsl(parsed_url.query, keep_blank_values=True, strict_parsing=True)
                    except Exception as exc:
                        raise CogmentError(f"Endpoint [{endpoint.url}]: {exc}")
                else:
                    query = []

                properties = {key: val for key, val in query}
                id = properties.get(ep.SERVICE_ID_PROPERTY_NAME)
                if not id:
                    raise CogmentError(f"A service path on a cogment endpoint must have an"
                                       f" '{ep.SERVICE_ID_PROPERTY_NAME}' query [{endpoint.url}]")

                service_id = int(id)
                dir_urls = await self.inquire_by_id(service_id)
                if len(dir_urls) == 0:
                    logger.error(f"No resource in directory with service id [{service_id}] [{endpoint.url}]")

                result = self._new_endpoint(endpoint, dir_urls)
                logger.debug(f"Inquired service endpoint result [{result.url}]")
                return result

            elif path == ep.ACTOR_PATH:
                dir_type = ServiceType.ACTOR
            elif path == ep.ENVIRONMENT_PATH:
                dir_type = ServiceType.ENVIRONMENT
            elif path == ep.DATALOG_PATH:
                dir_type = ServiceType.DATALOG
            elif path == ep.PRE_HOOK_PATH:
                dir_type = ServiceType.HOOK
            elif path == ep.LIFECYCLE_PATH:
                dir_type = ServiceType.LIFE_CYCLE
            elif path == ep.ACTOR_SERVICE_PATH:
                dir_type = ServiceType.CLIENT_ACTOR
            elif path == ep.DATASTORE_PATH:
                dir_type = ServiceType.DATASTORE
            elif path == ep.MODEL_REGISTRY_PATH:
                dir_type = ServiceType.MODEL_REG
            else:
                raise CogmentError(f"Unknown path for cogment endpoint [{endpoint.url}]")

        if parsed_url.query:
            try:
                query = urlpar.parse_qsl(parsed_url.query, keep_blank_values=True, strict_parsing=True)
            except Exception as exc:
                raise CogmentError(f"Endpoint [{endpoint.url}]: {exc}")
        else:
            query = []

        if properties is None:
            properties = {key: val for key, val in query}
        else:
            for key, val in query:
                if key in properties:
                    logger.warning(f"Key [{key}] in parameter 'properties'"
                                   f" will be overwriten with endpoint url query value [{val}]")
                properties[key] = val

        dir_urls = await self.inquire_by_type(dir_type, properties)
        if len(dir_urls) == 0:
            logger.error(f"No service in directory for [{endpoint.url}] with [{dir_type}] and [{properties}]")
        else:
            logger.debug(f"Inquired directory for [{endpoint.url}] with [{dir_type}] and [{properties}]")

        result_endpoint = self._new_endpoint(endpoint, dir_urls)
        logger.debug(f"Inquired endpoint result [{result_endpoint.url}]")

        return result_endpoint
